from csvw.dsv import reader

from linglit import util

CFG_PATH = util.CFG_PATH / 'langsci'
BIBTOOL_RSC = CFG_PATH / 'bibtool.rsc'
LNAME_MAP = {
    'Logoori': 'logo1258',
    'Līkpākpáln': 'konk1269',
    'Totoró Namtrik': 'toto1306',
    'Jóola Fóoñi': 'jola1263',
    'Mojeño Trinitario': 'trin1274',
    'Bùlì': 'buli1254',
    'Sereer-Siin': 'sere1260',  # 45
    'Fròʔò': 'tagw1240',  # 45
    'Siwi Berber': 'siwi1239',  # 41
    'Hoocąk': 'hoch1243',  # 40
    'Veraa': 'vera1241',  # 39
    'Early Vedic': '',  # 36
    'Greek, Attic': 'atti1240',  # 34
    'Late Modern Swedish': '',  # 30
    'Sembiran Balinese': '',  # 24
    'Beirut/Damascus': '',  # 24
    'Tsotsil': 'tzot1259',  # 23
    'Kakataibo': 'cash1251',  # 23
    'Bantu': '',  # 22
    'North Sámi': '',  # 22
    'Nganasan  (Avam)': 'avam1236',  # 21
    'Lycopolitan Coptic': 'lyco1237',  # 20
    'inglês': 'stan1293',  # 18
    'Greek, Classical|(': 'anci1242',  # 18
    'Greek, Homeric': 'anci1242',  # 17
    'Greek, Homeric|(': 'anci1242',  # 17
    'Greek, Cypriot': 'cypr1249',  # 17
    "K'abeena": 'alab1254',  # 16
    'francês': 'stan1290',  # 16
    'Luragooli': 'logo1258',  # 15
    'Rhonga': 'rong1268',  # 15
    'Sino-Japanese': '',  # 14
    'Hellenic': '',  # 14
    'Slavonic': 'chur1257',  # 13
    'Greek, Doric': 'dori1248',  # 13
    'Yixing Chinese': '',  # 13
    'Standard German': 'stan1295',  # 13
    'Allemand': 'stan1295',  # 12
    'Ioway, Otoe-Missouria': 'iowa1245',  # 12
    'Tanti Dargwa': '',  # 12
    'Singapore Malay': 'mala1479',  # 42
    'Early Modern Japanese': 'nucl1643',  # 23
    'Gulf Pidgin Arabic': 'pidg1248',  # 4
    'Nêlêmwa': 'kuma1276',  # 4
    'Övdalian': 'elfd1234',  # 15
    'Present-day Swedish': 'stan1279',  # 14
    'Älvdalen (Os)': 'elfd1234',  # 11
    'Sollerön (Os)': 'soll1234',  # 6
    'Orsa (Os)': 'orsa1234',  # 6
    'Nama-Damara': 'dama1270',  # 5
    'Nǀuuki': 'nuuu1241',  # 5
    r'\LangTurk': 'nucl1301',  # 5
    r'\LangTok': 'tokp1240',  # 4
    r'\LangVed': 'sans1269',  # 4
    r'\LangJap': 'nucl1643',  # 4
    r'\LangQue': 'quec1387',  #
    r'\LangMang': 'mang1381',  # 3
    r'\LangMand': 'mand1415',  # 3
    'Standard Greek': 'mode1248',  # 3
    'Överkalix (Kx)': 'arch1246',  # 3
    'Donno Sɔ': 'donn1239',  # 3
    'Ḥassāniyya': 'hass1238',  # 3
}


def iter_texfile_titles():
    yield from reader(CFG_PATH / 'texfile_titles.tsv', delimiter='\t', dicts=True)



from clldutils.clilib import PathType

from linglit import PROVIDERS


def add_provider(parser):
    parser.add_argument('provider', choices=list(PROVIDERS.keys()))
    parser.add_argument('dir', type=PathType(type='dir'))


def get_provider(args):
    return PROVIDERS[args.provider](args.dir)


def add_publication(parser):
    add_provider(parser)
    parser.add_argument('pubid')


def get_publication(args):
    return get_provider(args)[args.pubid]



import re
import logging
import functools

from clldutils import lgr
from pylatexenc import latexwalker, latex2text, macrospec
from pyigt.igt import NON_OVERT_ELEMENT

__all__ = ['simple_to_text', 'to_text', 'strip_tex_comment', 'iter_abbreviations']

# Replace \<key> with <value> for the following macros:
SIMPLE_MACROS = {
    'ANA': 'ANA',
    'Tilde': '~',
    'tld': '~',
    'oneS': '1S',
    'twoS': '2S',
    'oneP': '1P',
    'twoP': '2P',
    'CL': 'CL',
    'ob': '[',
    'cb': ']',
    'op': '(',
    'cp': ')',
    'db': ' [',
    'R': 'r',
    'Nonpast': 'NONPAST',
    'redp': '~',
    'Stpl': '2|3PL',
    'Masc': 'MASC',
    'PSP': 'PSP',
    'Fsg': '1SG',
    'Fpl': '1PL',
    'ATT': 'ATT',
    'deff': 'DEFF',
    'Ssg': '2SG',
    'Stsg': '2|3SG',
    'Third': '3',
    'Tsg': '3SG',
    'IFV': 'IFV',
    'Second': '2',
    'HOR': 'HOR',
    'pp': 'PP',
    'Aand': '&',
    'IO': 'IO',
    'Io': 'IO',
    'LINK': 'LINK',
    'flobv': '→3',
    'fl': '→',
    'Venit': 'VENT',
    'Ndu': 'NDU',
    'PN': 'PN',
    'slash': '/',
    'DEP': 'DEP',
    'USSmaller': '<',
    'USGreater': '>',
    'defn': 'DEFN',
    'Char': 'CHAR',
    'DO': 'DO',
    'Rpst': 'RPST',
    'Ext': 'EXT',
    'EXT': 'EXT',
    'shin': 'ʃ',
    'shinB': 'ʃ',
    'alef': 'ʔ',
    'alefB': 'ʔ',
    'ayin': 'ʕ',
    'ayinB': 'ʕ',
    'het': 'ħ',
    'hetB': 'ħ',
    'PRON': 'PRON',
    'NOUN': 'NOUN',
    'CONJ': 'CONJ',
    'glossF': 'F',
    'glossNeg': 'NEG',
    'glsg': 'SG',
    'sq': '~',
    'INSTR': 'INSTR',
    'POT': 'POT',
    'PLU': 'PLU',
    'RETRO': 'RETRO',
    'CONTR': 'CONTR',
    'USOParen': '(',
    'USCParen': ')',
    'conn': 'CONN',
    'At': 'AT',
    'textepsilon': 'ɛ',
    'textopeno': 'ɔ',
    'textbeltl': 'ɬ',
    'textless': '<',
    'textgreater': '>',
    'textltailn': 'ɲ',
    'textquotedbl': '"',
    'textsci': 'ɪ',
    'IDPH': 'IDPH',
    'ID': 'ID',
    'Priv': 'PRIV',  # privative case
    'MOD': 'MOD',
    'Aor': 'AOR',
    'glossINF': 'INF',
    'Subj': 'SUBJ',
    'abi': 'ABI',
    'hest': 'HEST',
    'quant': 'QUANT',
    'RECIP': 'RECIP',
    'USEmptySet': '∅',
    'Obl': 'OBL',
    'Indic': 'INDIC',
    'Dei': 'DEI',
    'Expl': 'EXPL',
    'Sdu': 'SDU',
    'Mid': 'MID',
    'Npst': 'NPST',
    'Nom': 'NOM',
    'Nw': 'NW',
    'Act': 'ACT',
    'textgamma': 'ɣ',
    'textperiodcentered': '·',
    'cras': 'CRAS',
    'PRIOR': 'PRIOR',
    'SIM': 'SIM',
    'Lnk': 'LNK',
    'Av': 'AV',
    'Ppp': 'PPP',
    'Futimp': 'FUTIMP',
    'deter': 'DETER',
    'mut': 'MUT',
    'NC': 'NC',
    'textglotstop': 'ʔ',
    'textraiseglotstop': 'ʔ',
    'Intj': 'INTJ',
    'Prt': 'PRT',
    'NEUT': 'NEUT',
    'PF': 'PF',
    'z': '',
    'expl': 'EXPL',
    'Emph': 'EMPH',
    'Hab': 'HAB',
    'Gam': 'GAM',
    'Vc': 'VC',
    'Stnsg': '2|3NSG',
    'Stdu': '2|3DU',
    'Only': 'ONLY',
    'Imn': 'IMN',
    'Rs': 'RS',
    'Recog': 'RECOG',
    'Iam': 'IAM',
    'ds': 'ꜜ',
    'Prop': 'PROP',
    'Betaone': 'Β1',
    'Betatwo': 'Β2',
    'Appr': 'APPR',
    'AP': 'AP',
    'Bet': 'Β',
    'Pot': 'POT',
    'Imm': 'IMM',
    'Immpst': 'IPST',
    'Fnsg': '1NSG',
    'Dim': 'DIM',
    'Med': 'MED',
    'Stat': 'STAT',
    'Zero': NON_OVERT_ELEMENT,
    'Fdu': '1DU',
    'Lk': 'LK',
    'Alph': 'Α',
    'reln': 'RELN',
    'postp': 'POSTP',
    'dxm': 'DXM',
    'ideo': 'IDEO',
    'IDEO': 'IDEO',
    'ident': 'IDENT',
    'psg': 'PSG',
    'mod': 'MOD',
    'itr': 'ITR',
    'hab': 'HAB',
    'egr': 'EGR',
    'qual': 'QUAL',
    'la': 'lá',
    'baru': 'ʉ',
    'baruL': 'ʉ̀',
    'baruH': 'ʉ́',
    'ep': 'ε',
    'epL': 'ὲ',
    'epH': 'έ',
    'schwa': 'ə',
    'schwaH': 'ə́',
    'schwaL': 'ə̀',
    'mbuHL': 'mbʉ́ʉ̀',
    'mbuLH': 'mbʉ̀ʉ́',
    'mbuL': 'mb̀ʉ̀',
    'oo': 'ɔ',
    'ooH': 'ɔ́',
    'ooL': 'ɔ̀',
    'nda': 'ndà',
    'longrule': '_',
    'cmpr': 'CMPR',
    'thestrong': 'the',
    'theweak': 'the',
    'emphat': 'EMPH',
    'pstpunc': 'PSTPUNC',
    'relativ': 'REL',
    'cert': 'CERT',
    'att': 'ATT',
    'antip': 'ANTIP',
    'prep': 'PREP',
    'partic': 'PART',
    'antgen': 'ANTGEN',
    'am': 'AM',
    'asp': 'ASP',
    'ideoph': 'IDEOPH',
    'prosp': 'PROSP',
    'PROSP': 'PROSP',
    'pstrem': 'PSTREM',
    'fem': 'F',
    'mas': 'M',
    'faff': 'FAFF',
    'intens': 'INTENS',
    ':': 'ː',
    'fin': 'FIN',
    'conj': 'CONJ',
    'allo': 'ALLO',
    'ines': 'INES',
    'icml': 'ICML',
    'tns': 'TNS',
    'uop': 'UOP',
    'Hyp': 'HYP',
    'dsbj': 'DSBJ',
    'topic': 'TOP',
    'ctopic': "CTOPIC",
    'dir': 'DIR',
    'cs': 'CS',
    'question': 'Q',
    'glopa': 'OPA',
    'glme': 'TOP',
    'glte': 'NFIN',
    'glta': 'NFIN',
    'Cl': 'CL',
    'Rci': 'RCI',
    'Spl': '2PL',
    'Sm': 'SM',
    'Fv': 'FV',
    'Aug': 'AUG',
    'Dj': 'DJ',
    'Om': 'OM',
    "Adlzr": "ADLZR",  # adjectivaliser
    'glossQ': 'Q',
    'Depend': 'DEPEND',
    'Simp': 'SIMP',
    'Vn': 'VN',
    'Asp': 'ASP',
    'Invol': 'INVOL',
    'Volition': 'volition',
    'sVP': 'VP',
    'part': 'PARTICLE',
    'prag': 'PRAG',
    'pda': 'PDA',
    'Recip': 'REC',
    'persm': 'PM',
    'complx': 'COMPLX',
    'defsc': 'DEF',
    'transitiv': 'TR',
    'detr': 'DETR',
    'epen': 'EPEN',
    'agnm': 'AGNM',
    'ssbj': 'SSBJ',
    'rdp': 'RDP',
    'NR': 'NR',
    'con': 'CON',
    'Verb': 'VBLZ',
    'instr': 'INS',
    'firstperson': '1',
    'secondperson': '2',
    'thirdperson': '3',
    'an': 'AN',
    'rep': 'REP',
    'illa': 'ILLA',
    'topi': 'TOP',
    'lptcp': 'l_PTCP',
    'masc': 'M',
    'negation': 'NEG',
    'object': 'O',
    'impr': 'IMPR',
    'add': 'ADD',
    'andat': 'AND',  # andative
    'Andat': 'AND',  # andative
    'nsg': 'NSG',  # non-singular
    'Nsg': 'NSG',
    'Jj': 'J',
    'Lv': 'LV',
    'Lat': 'LAT',
    'Tns': 'TNS',
    'Etc': 'ETC',
    'Cmpr': 'CMPR',
    'AUTOCAUS': 'AUTOCAUS',
    'HAB': 'HAB',
    'Iii': 'III',
    'Cont': 'CONT',
    'Int': 'INT',
    'Prov': 'PROV',
    'Perm': 'PERM',
    'Cntr': 'CNTR',
    'Dv': 'DV',  # default vowel
    'TNS': 'TNS',
    'ASP': 'ASP',
    'glinf': 'INF',
    'glgen': 'GEN',
    'glprog': 'PROG',
    'Rem': 'REM',  # remote
    'Ti': 'TI',
    'Inv': 'INV',
    'Ss': 'SS',
    'Su': 'S',
    'Tpl': '3PL',
    'Desc': 'DESC',
    'Red': 'RED',
    'Iter': 'ITER',
    'Th': 'TH',
    'Vblz': 'VBLZ',
    'Detr': 'DETR',
    'Nth': 'NTH',
    'zero': NON_OVERT_ELEMENT,
    'sprl': 'SPRL',
    'wk': 'WK',
    'Ints': 'INTS',
    'Ds': 'DS',  # different subject
    'First': '1',
    'GH': 'ɣ',
    'vir': 'VIR',
    'nvir': 'N_VIR',
    'rls': 'RLS',
    'bg': 'BG',
    'stat': 'STAT',
    'dep': 'DEP',
    'punc': 'PUNC',
    'neu': 'N',
    'pstcont': 'PSTCONT',
    'polite': 'POL',
    'h': "'",
    'eer': 'ε',
    "textrtaild": "ɖ",
    "D": "ɖ",  # \textrtaild} %for retroflex d
    "F": "ʄ",  # \texthtbardotlessj}
    "oor": "ɔ",  # \textopeno}  % for ooor
    "Ng": "N",  # \textipa{N}}  %for NG
    "shw": "ə",  # \textschwa}  %for Shwa
    "gy": "ʤ",  # \textdyoghlig}  %for GY
    "ssh": "ʃ",  # \textesh} %ssh
    "tts": "ʧ",  # \textteshlig} %ts
    "zzh": "ʒ",  # \textyogh} %zzh
    "ny": "ɲ",  # \textltailn}
    "ph": "ɸ",  # \textphi}  %phi
    "lw": "`",  # \`} %LOW
    # "ds": "",  # \textdownstep} %DOWNSTEP
    "ns": "~",  # \~} %NAS
    "syl": "σ",  # $\sigma$} %SYLL
    "mo": "μ",  # $\mu$} %MORA
    "rnk": "≫",  # $\gg$} %>>
    "textcrh": "ħ",
    "rs": "\u030c",  # \v}
    'subj': "SUBJ",
    'PREP': 'PREP',
    'PAST': 'PAST',
    'glossM': 'M',
    'PRES': 'PRES',
    'NUM': 'NUM',
    'NOUNPROPER': 'NOUNPROPER',
    'ESS': 'ESS',
    'ESSIVE': 'ESSIVE',
    'ADESS': 'ADESS',
    'Redup': 'REDUP',
    'Anim': 'ANIM',
    'Temp': 'TEMP',
    'Snsg': '2NSG',
    'Pos': 'POS',  # positional verb stem
    'ᵓ': 'ᵓ',
    '°': '°',
    "ᵃ": "ᵃ",  # \textsf{{\hspace{.5pt}ᵃ}}}
    "ᵋ": "ᵋ",  # \textsf{{\hspace{.5pt}ᵋ}}}
    "ᵉ": "ᵉ",  # \textsf{{\hspace{.1pt}ᵉ}}}
    "ⁱ": "ⁱ",  # {\raisebox{-.65mm": "",  #{\textsuperscript{\sffamily\scriptsize i}}}}}
    "ᴵ": "ᴵ",  # \textsf{{\hspace{.5pt}ᴵ}}}
    "ᶤ": "ᶤ",  # \textsf{{\hspace{.5pt}ᶤ}}}
    "ᵒ": "ᵒ",  # \textsf{{\hspace{.5pt}ᵒ}}}
    "ᵘ": "ᵘ",  # \textsf{{\hspace{.5pt}ᵘ}}}
    "ᶶ": "ᶶ",  # \textsf{{\hspace{.5pt}ᶶ}}}
    "ꜜ": "ꜜ",  # \textsf{ꜜ\hspace*{-2pt}}}
    'supperl': 'SUPPERL',
    'supp': 'SUP',
    'definite': 'DEF',
    "PART": "PART",
    "PRV": "PRV",
    "STR": "STR",
    "INTENS": "INTENS",
    "HUMAN": "HUMAN",
    "NONHUMAN": "NONHUMAN",
    "ASSERT": "ASSERT",
    "IMPV": "IMPV",
    "IMPF": "IMPF",
    "INT": "INTR",
    "POSTP": "POSTP",
    'MAS': 'M',
    "Fosg": "4SG",  # {\gls{4}\gls{sg}}
    "Fopl": "4PL",  # {\gls{4}\gls{pl}}
    "Abl": "ABL",  # {\gls{abl}}
    "Abs": "ABS",  # {\gls{abs}}
    "Acc": "ACC",  # {\gls{acc}}
    "Ade": "ADE",  # {\gls{ade}}
    "Adj": "ADJ",  # {\gls{adj}}
    "Adv": "ADV",  # {\gls{adv}}
    "Agr": "AGR",  # {\gls{agr}}
    'ALL': 'ALL',
    "All": "ALL",  # {\gls{all}}
    "Anom": "ADNOM",  # {\gls{adnom}}
    "Ant": "ANT",  # {\gls{ant}}
    "Antip": "ANTIP",  # {\gls{antip}}
    "An": "AN",  # {\gls{an}}
    "Appl": "APPL",  # {\gls{appl}}
    "Apud": "APUD",  # {\gls{apud}}
    "Art": "ART",  # {\gls{art}}
    "Assoc": "ASSOC",  # {\gls{assoc}}
    "Aud": "AUD",  # {\gls{aud}}
    "Aux": "AUX",  # {\gls{aux}}
    "Bb": "B",  # {\gls{B}}
    "Ben": "BEN",  # {\gls{ben}}
    "BrP": "BRP",  # {\gls{BrP}}
    "Caus": "CAUS",  # {\gls{caus}}
    "Cci": "CCI",  # {\gls{cci}}
    "Cf": "CF",  # {\gls{cf}}
    "Cj": "CJ",  # {\gls{cj}}
    "Clf": "CLF",  # {\gls{clf}}
    "Cmpl": "CMPL",  # {\gls{cmpl}}
    "Cnj": "CNJ",  # {\gls{cnj}}
    "Comp": "COMP",  # {\gls{comp}}
    "Compl": "COMPL",  # {\gls{compl}}
    "Com": "COM",  # {\gls{com}}
    "Cond": "COND",  # {\gls{cond}}
    "Conn": "CONN",  # {\gls{conn}}
    "Cop": "COP",  # {\gls{cop}}
    "Cr": "CR",  # {\gls{cr}}
    "Cvb": "CVB",  # {\gls{cvb}}
    "Dat": "DAT",  # {\gls{dat}}
    "Decl": "DECL",  # {\gls{decl}}
    "Def": "DEF",  # {\gls{def}}
    "Dem": "DEM",  # {\gls{dem}}
    "Dep": "DEP",  # {\gls{dep}}
    "Det": "DET",  # {\gls{det}}
    "Dir": "DIR",  # {\gls{dir}}
    "Disp": "DISP",  # {\gls{disp}}
    "Distr": "DISTR",  # {\gls{distr}}
    "Dist": "DIST",  # {\gls{dist}}
    "Dom": "DOM",  # {\gls{dom}}
    "Do": "DO",  # {\gls{do}}
    "Dptcl": "DPTCL",  # {\gls{dptcl}}
    "Dr": "DR",  # {\gls{dr}}
    "Dur": "DUR",  # {\gls{dur}}
    "Du": "DU",  # {\gls{du}}
    "Ela": "ELA",  # {\gls{ela}}
    "Ep": "EP",  # {\gls{ep}}
    "Erg": "ERG",  # {\gls{erg}}
    "Es": "ES",  # {\gls{es}}
    "EuP": "EUP",  # {\gls{EuP}}
    "Evid": "EVID",  # {\gls{evid}}
    "Excl": "EXCL",  # {\gls{excl}}
    "Foc": "FOC",  # {\gls{foc}}
    "Fs": "FS",  # {\gls{fs}}
    "Fut": "FUT",  # {\gls{fut}}
    "Gen": "GEN",  # {\gls{gen}}
    "Genc": "GENC",  # {\gls{genc}}
    "Ger": "GER",  # {\gls{ger}}
    "Gg": "G_grade_(Muskogean)",  # {\gls{g}-grade (Muskogean)}
    "Ill": "ILL",  # {\gls{ill}}
    "Impass": "IMPASS",  # {\gls{impass}}
    "Imp": "IMP",  # {\gls{imp}}
    "Inch": "INCH",  # {\gls{inch}}
    'INCH': 'INCH',
    "Incl": "INCL",  # {\gls{incl}}
    "Incmpl": "INCMPL",  # {\gls{incmpl}}
    "Indef": "INDEF",  # {\gls{indef}}
    "Indir": "INDIR",  # {\gls{indir}}
    "Ind": "IND",  # {\gls{ind}}
    "Ine": "INE",  # {\gls{ine}}
    "Inf": "INF",  # {\gls{inf}}
    'inf': 'INF',
    "Infr": "INFR",  # {\gls{infr}}
    "Ins": "INS",  # {\gls{ins}}
    "Intr": "INTR",  # {\gls{intr}}
    "Ipfv": "IPFV",  # {\gls{ipfv}}
    "Ipf": "IPF",  # {\gls{ipf}}
    "Irr": "IRR",  # {\gls{irr}}
    "Ll": "L",  # {\gls{l}}
    "Loc": "LOC",  # {\gls{loc}}
    "M": "M",  # {\gls{m}}
    "Mnom": "MNOM",  # {\gls{mnom}}
    "Mr": "MR",  # {\gls{mr}}
    "Mut": "MUT",  # {\gls{mut}}
    "glossN": "N",  # {\gls{n}}
    "Nact": "NACT",  # {\gls{nact}}
    "Ncm": "NCM",  # {\gls{ncm}}
    "Neg": "NEG",  # {\gls{neg}}
    "Nfin": "NFIN",  # {\gls{nfin}}
    "Nmlz": "NMLZ",  # {\gls{nmlz}}
    "Num": "NUM",  # {\gls{num}}
    "Obj": "OBJ",  # {\gls{obj}}
    "Obv": "OBJV",  # {\gls{obv}}
    "Ov": "OV",  # {\gls{ov}}
    "Part": "PART",  # {\gls{part}}
    "Pass": "PASS",  # {\gls{pass}}
    "Pauc": "PAUC",  # {\gls{pauc}}
    "Pfv": "PFV",  # {\gls{pfv}}
    "Pg": "PG",  # {\gls{pg}}
    "Pip": "PIP",  # {\gls{pip}}
    "Pl": "PL",  # {\gls{pl}}
    "Pn": "PN",  # {\gls{pn}}
    "Poss": "POSS",  # {\gls{poss}}
    "Prdst": "PRDST",  # {\gls{prdst}}
    "Pred": "PRED",  # {\gls{pred}}
    "Prf": "PRF",  # {\gls{prf}}
    "Prep": "PREP",  # {\gls{prep}}
    "Prog": "PROG",  # {\gls{prog}}
    "Proh": "PROH",  # {\gls{proh}}
    "Pron": "PRON",  # {\gls{pron}}
    "Propn": "PROPN",  # {\gls{propn}}
    "Pros": "PROS",  # {\gls{pros}}
    "Prox": "PROX",  # {\gls{prox}}
    "Prs": "PRS",  # {\gls{prs}}
    "Prv": "PRV",  # {\gls{prv}}
    "Pst": "PST",  # {\gls{pst}}
    "Ptcl": "PTCL",  # {\gls{ptcl}}
    "Ptcp": "PTCP",  # {\gls{ptcp}}
    "ptcp": "PTCP",
    "Purp": "PURP",  # {\gls{purp}}
    "Pv": "PV",  # {\gls{pv}}
    "Quot": "QUOT",  # {\gls{quot}}
    "Real": "REAL",  # {\gls{real}}
    "Recp": "RECP",  # {\gls{recp}}
    "Refl": "REFL",  # {\gls{refl}}
    "Rel": "REL",  # {\gls{rel}}
    "Rep": "REP",  # {\gls{rep}}
    "Res": "RES",  # {\gls{res}}
    "Rg": "RG",  # {\gls{rg}}
    "Rm": "RM",  # {\gls{rm}}
    "Rpt": "RPT",  # {\gls{rpt}}
    "Sbjv": "SBJV",  # {\gls{sbjv}}
    "Sbj": "SBJ",  # {\gls{sbj}}
    "Secun": "SEC",  # {\gls{sec}}
    "Sg": "SG",  # {\gls{sg}}
    "Sim": "SIM",  # {\gls{sim}}
    "Sr": "SR",  # {\gls{sr}}
    "Subl": "SUBL",  # {\gls{subl}}
    "Supe": "SUPE",  # {\gls{supe}}
    "Supl": "SUPL",  # {\gls{supl}}
    "Ta": "TA",  # {\gls{ta}}
    "Term": "TERM",  # {\gls{term}}
    "Topic": "TOP",  # {\gls{top}}
    "Transl": "TRANS",  # {\gls{transl}}
    "Tr": "TR",  # {\gls{tr}}
    "Tto": "TT",  # {\gls{tt}}
    "Ut": "UT",  # {\gls{ut}}
    "Uu": "U",  # {\gls{u}}
    "Vm": "VM",  # {\gls{vm}}
    "Voc": "VOC",  # {\gls{voc}}
    "Vol": "VOL",  # {\gls{vol}}
    "Yn": "Y.N",  # {\gls{y.n}}
    "Aa": "A",  # {\gls{a}}
    "Pp": "P",  # {\gls{p}}
    "Rr": "R",  # {\gls{r}}
    "Tt": "T",  # {\gls{t}}
    "Ig": "I",  # {\gls{i}}
    "Ii": "II",  # {\gls{ii}}
    "Iv": "IV",  # {\gls{iv}}
    "SCC": "SC3",  # {\gls{SC3}}
    "SC": "SC2",  # {\gls{SC2}}
    'SUB': 'SUB',
    "nom": "NOM",  # {{\sc nom}}
    "gen": "GEN",  # {{\sc gen}}
    "dat": "DAT",  # {{\sc dat}}
    "acc": "ACC",  # {{\sc acc}}
    "abs": "ABS",  # {{\sc abs}}
    "erg": "ERG",  # {{\sc erg}}
    "voc": "VOC",  # {{\sc voc}}
    "poss": "POSS",  # {{\sc poss}}
    "pl": "PL",  # {{\sc pl}}
    "sg": "SG",  # {{\sc sg}}
    "thirdsg": "3SG",  # {3{\sc sg}}
    "ali": "ALI",  # {{\sc ali}}
    "appl": "APPL",  # {{\sc appl}}
    "augv": "AUGV",  # {{\sc augv}}
    "aux": "AUX",  # {{\sc aux}}
    "ben": "BEN",  # {{\sc ben}}
    "caus": "CAUS",  # {{\sc caus}}
    "clf": "CLF",  # {{\sc clf}}
    "coll": "COLL",  # {{\sc coll}}
    "com": "COM",  # {{\sc com}}
    "comp": "COMP",  # {{\sc comp}}
    "compl": "COMPL",  # {{\sc compl}}
    "cond": "COND",  # {{\sc cond}}
    "cop": "COP",  # {{\sc cop}}
    "cvb": "CVB",  # {{\sc cvb}}
    "decl": "DECL",  # {{\sc decl}}
    "dem": "DEM",  # {{\sc dem}}
    "dist": "DIST",  # {{\sc dist}}
    "distr": "DISTR",  # {{\sc distr}}
    "dscn": "DSCN",  # {{\sc dscn}}
    "cn": "CN",
    "du": "DU",  # {{\sc du}}
    "dubt": "DUBT",  # {{\sc dubt}}
    "dur": "DUR",  # {{\sc dur}}
    "dyn": "DYN",  # {{\sc dyn}}
    "ego": "EGO",  # {{\sc ego}}
    "excl": "EXCL",  # {{\sc excl}}
    "foc": "FOC",  # {{\sc foc}}
    "fut": "FUT",  # {{\sc fut}}
    "imp": "IMP",  # {{\sc imp}}
    "incep": "INCEP",  # {{\sc incep}}
    "ind": "IND",  # {{\sc ind}}
    "indf": "INDF",  # {{\sc indf}}
    "intr": "INTR",  # {{\sc intr}}
    "ipfv": "IPFV",  # {{\sc ipfv}}
    "irr": "IRR",  # {{\sc irr}}
    "lin": "LIN",  # {{\sc lin}}
    "loc": "LOC",  # {{\sc loc}}
    "modal": "MODAL",  # {{\sc modal}}
    "Non": "N",  # {{\sc n}}
    "nmlz": "NMLZ",  # {{\sc nmlz}}
    "nts": "NTS",  # {{\sc nts}}
    "obj": "OBJ",  # {{\sc obj}}
    "obl": "OBL",  # {{\sc obl}}
    "opt": "OPT",  # {{\sc opt}}
    "pass": "PASS",  # {{\sc pass}}
    "propnoun": "PN",  # {{\sc pn}}
    "pronoun": "PRO",  # {{\sc pro}}
    "pred": "PRED",  # {{\sc pred}}
    "prf": "PRF",  # {{\sc prf}}
    "prog": "PROG",  # {{\sc prog}}
    "prs": "PRS",  # {{\sc prs}}
    "prtv": "PRTV",  # {{\sc prtv}}
    "proh": "PROH",  # {{\sc proh}}
    "propr": "PROPR",  # {{\sc propr}}
    "prox": "PROX",  # {{\sc prox}}
    "pst": "PST",  # {{\sc pst}}
    "purp": "PURP",  # {{\sc purp}}
    "pfv": "PFV",  # {{\sc pfv}}
    "pvs": "PVS",  # {{\sc pvs}}
    "reas": "REAS",  # {{\sc reas}}
    "refl": "REFL",  # {{\sc refl}}
    "result": "RES",  # {{\sc res}}
    "sbj": "SBJ",  # {{\sc sbj}}
    "semb": "SEMB",  # {{\sc semb}}
    "simult": "SIM",  # {{\sc sim}}
    "them": "THEM",  # {{\sc them}}
    "tmp": "TMP",  # {{\sc tmp}}
    "ventiv": "VEN",  # {{\sc ven}}
    "ELAT": "ELAT",
    "E": "E",  # epenthetic morpheme
    "ATTRs": "ATTR",  # {{\Sc{attr}}}%shortcut for ATTR in small caps
    "PREDs": "PRED",  # {{\Sc{pred}}}%shortcut for PRED in small caps
    "SGs": "SG",  # {{\Sc{sg}}}%shortcut for SG in small caps
    "DUs": "DU",  # {{\Sc{du}}}%shortcut for DU in small caps
    "PLs": "PL",  # {{\Sc{pl}}}%shortcut for PL in small caps
    "NOMs": "NOM",  # {{\Sc{nom}}}%shortcut for NOM in small caps
    "ACCs": "ACC",  # {{\Sc{acc}}}%shortcut for ACC in small caps
    "GENs": "GEN",  # {{\Sc{gen}}}%shortcut for GEN in small caps
    "ILLs": "ILL",  # {{\Sc{ill}}}%shortcut for ILL in small caps
    "INESSs": "INESS",  # {{\Sc{iness}}}%shortcut for INESS in small caps
    "ELATs": "ELAT",  # {{\Sc{elat}}}%shortcut for ELAT in small caps
    "COMs": "COM",  # {{\Sc{com}}}%shortcut for COM in small caps
    "ABESSs": "ABESS",  # {{\Sc{abess}}}%shortcut for ABESS in small caps
    "ESSs": "ESS",  # {{\Sc{ess}}}%shortcut for ESS in small caps
    "PROXs": "PROX",  # {{\Sc{prox}}}%shortcut for PROX in small caps
    "DISTs": "DIST",  # {{\Sc{dist}}}%shortcut for DIST in small caps
    "RMTs": "RMT",  # {{\Sc{rmt}}}%shortcut for RMT in small caps
    "REFLs": "REFL",  # {{\Sc{refl}}}%shortcut for REFL in small caps
    "PRSs": "PRS",  # {{\Sc{prs}}}%shortcut for PRS in small caps
    "PSTs": "PST",  # {{\Sc{pst}}}%shortcut for PST in small caps
    "IMPs": "IMP",  # {{\Sc{imp}}}%shortcut for IMP in small caps
    "POTs": "POT",  # {{\Sc{pot}}}%shortcut for POT in small caps
    "PROGs": "PROG",  # {{\Sc{prog}}}%shortcut for PROG in small caps
    "PRFs": "PRF",  # {{\Sc{prf}}}%shortcut for PRF in small caps
    "INFs": "INF",  # {{\Sc{inf}}}%shortcut for INF in small caps
    "NEGs": "NEG",  # {{\Sc{neg}}}%shortcut for NEG in small caps
    "CONNEGs": "CONNEG",  # {{\Sc{conneg}}}%shortcut for CONNEG in small caps
    "BS": "<backslash>",
    'PRG': 'PRG',
    "pafr": "š",
    "paaf": "č",  # voiceless post-alveolar affricate
    "vuvfr": "ʁ",  # voiced uvular fricative
    "uvfr": "χ",  # voiceless uvular fricative
    "phfr": "ħ",  # voiceless pharyngeal fricative
    "eppl": "ʡ",  # epiglottal plosive
    "glpl": "ʔ",  # glottal stop
    "glst": "ʔ",  # alias for glottal stop
    "ej": "’",  # ejective
    "lab": "ʷ",  # labialized
    "pha": "ˤ",  # pharyngealized
    "lmk": "ː",
    "tet": "ṭ",
    "dao": "ố",
    "daa": "ấ",
    "dae": "ế",
    "daob": "ồ",
    "daab": "ầ",
    "daeb": "ề",
    'ts': "ts",
    'textdownstep': 'ꜜ',
    'Simil': 'SIMIL',  # similative
    'SSS': '3S',
    'ingr': 'INGR',
}

logging.getLogger('pylatexenc.latexwalker').setLevel(logging.WARNING)

#
# Define macros for the *parser*
#
macros = [
    macrospec.MacroSpec("footnotetext", "{"),
    macrospec.MacroSpec("footnote", "{"),
    macrospec.MacroSpec("japhdoi", "{"),
    macrospec.MacroSpec("textup", "{"),
    macrospec.MacroSpec("textupsc", "{"),
    macrospec.MacroSpec("glossfeat", "{"),
    macrospec.MacroSpec("langinfo", "{{{"),
    macrospec.MacroSpec("tss", "{"),
    macrospec.MacroSpec("ili", "{"),
    macrospec.MacroSpec("ilt", "{"),
    macrospec.MacroSpec("il", "{"),
    macrospec.MacroSpec("is", "{"),
    macrospec.MacroSpec("ist", "{"),
    macrospec.MacroSpec("ia", "{"),
    macrospec.MacroSpec("ix", "{"),
    macrospec.MacroSpec("ux", "{"),
    macrospec.MacroSpec("ref", "{"),
    macrospec.MacroSpec("llap", "{"),
    macrospec.MacroSpec("textsc", "{"),
    macrospec.MacroSpec("Sc", "{"),
    macrospec.MacroSpec("tsc", "{"),
    macrospec.MacroSpec("gsc", "{"),
    macrospec.MacroSpec("ig", "{"),
    macrospec.MacroSpec("linieb", "{{"),
    macrospec.MacroSpec("ulp", "{{"),
    macrospec.MacroSpec("ulg", "{{"),
    macrospec.MacroSpec("japhug", "{{"),
    macrospec.MacroSpec("gloss", "{"),
    macrospec.MacroSpec("REF", "{"),
    macrospec.MacroSpec("mc", "{"),
    macrospec.MacroSpec("particle", "{"),
    macrospec.MacroSpec("jambox", "[{"),
    macrospec.MacroSpec("mbox", "{"),
    macrospec.MacroSpec("blockcquote", "{{"),
    macrospec.MacroSpec("scite", "{{"),
    macrospec.MacroSpec("fatcit", "{{"),
    macrospec.MacroSpec("fatcitNP", "{{"),
    macrospec.MacroSpec("nocite", "{"),
    macrospec.MacroSpec("possessivecite", "[[{"),
    macrospec.MacroSpec("pgcitet", "[[{"),
    macrospec.MacroSpec("posscite", "[[{"),
    macrospec.MacroSpec("posscitet", "[[{"),
    macrospec.MacroSpec("posscitealt", "[[{"),
    macrospec.MacroSpec("namecite", "[[{"),
    macrospec.MacroSpec("Textcite", "[[{"),
    macrospec.MacroSpec("textcite", "[[{"),
    macrospec.MacroSpec("textcites", "[[{"),
    macrospec.MacroSpec("parencite", "[{"),
    macrospec.MacroSpec("textcquote", "[{"),
    macrospec.MacroSpec("href", "{{"),
    macrospec.MacroSpec("dline", "{{{"),
    macrospec.MacroSpec("autocite", "[[{"),
    macrospec.MacroSpec("autocites", "[[{"),
    macrospec.MacroSpec("cite", "[[{"),
    # FIXME: 17 Corpus and CorpusE
]
for k in SIMPLE_MACROS:
    macros.append(macrospec.MacroSpec(k, ""))
for abbr in lgr.ABBRS:
    if abbr:
        if abbr not in SIMPLE_MACROS:
            macros.append(macrospec.MacroSpec(abbr, ""))
        if abbr.lower() not in SIMPLE_MACROS:
            macros.append(macrospec.MacroSpec(abbr.lower(), ""))
        if len(abbr) > 1:
            if abbr.capitalize() not in SIMPLE_MACROS:
                macros.append(macrospec.MacroSpec(abbr.capitalize(), ""))

lw_context_db = latexwalker.get_default_latex_context_db()
lw_context_db.add_context_category('gll', prepend=True, macros=macros[:])

simple_parser_context_db = latexwalker.get_default_latex_context_db()
simple_parser_context_db.add_context_category(
    'simple',
    prepend=True,
    macros=[macrospec.MacroSpec("href", "{{")])


#
# Implement macros for the *conversion to text*
#

def uppercase_arg(n, l2tobj):
    if n.nodeargd:
        return l2tobj.nodelist_to_text([n.nodeargd.argnlist[0]]).upper()
    return ''


def dot_uppercase_arg(n, l2tobj):
    return '.' + l2tobj.nodelist_to_text([n.nodeargd.argnlist[0]]).upper()


def footnote(n, l2tobj):
    return "§fn§{}§/fn§".format(l2tobj.nodelist_to_text([n.nodeargd.argnlist[0]]))


def japhdoi(n, l2tobj):
    return '<a href="https://doi.org/10.24397/pangloss-{}"></fn>'.format(
        l2tobj.nodelist_to_text([n.nodeargd.argnlist[0]]))


def lastarg(n, l2tobj):
    return l2tobj.nodelist_to_text([n.nodeargd.argnlist[-1]])


def firstarg(n, l2tobj):
    if n.nodeargd:
        return l2tobj.nodelist_to_text([n.nodeargd.argnlist[0]])
    return ''


def secondarg(n, l2tobj):
    return l2tobj.nodelist_to_text([n.nodeargd.argnlist[1]])


def repl(abbr, *args):
    return abbr


def japhug(n, l2tobj):
    return "{} [{}]".format(
        l2tobj.nodelist_to_text([n.nodeargd.argnlist[0]]),
        l2tobj.nodelist_to_text([n.nodeargd.argnlist[1]]),
    )


def href(n, l2tobj):
    return "[{}]({})".format(
        l2tobj.nodelist_to_text([n.nodeargd.argnlist[1]]),
        l2tobj.nodelist_to_text([n.nodeargd.argnlist[0]]),
    )


def scite(n, l2tobj):
    return '<cit page="">{}</cit>'.format(l2tobj.nodelist_to_text([n.nodeargd.argnlist[1]]))


def fatcit(n, l2tobj):
    return '<cit page="{}">{}</cit>'.format(
        l2tobj.nodelist_to_text([n.nodeargd.argnlist[1]]),
        l2tobj.nodelist_to_text([n.nodeargd.argnlist[0]]))


def _get_optional_arg(node, default, l2tobj):
    """Helper that returns the `node` converted to text, or `default`
    if the node is `None` (e.g. an optional argument that was not
    specified)"""
    if node is None:
        return default
    return l2tobj.nodelist_to_text([node])


def cite(n, l2tobj):
    """Get the text replacement for the macro \\cite...[page]{bibkey}"""
    if not n.nodeargd:
        # n.nodeargd can be empty if e.g. \putinquotes was a single
        # token passed as an argument to a macro,
        # e.g. \newcommand\putinquotes...
        return ''
    page = ''
    if len(n.nodeargd.argnlist) > 1:
        page = _get_optional_arg(n.nodeargd.argnlist[0], '', l2tobj)
    key = l2tobj.nodelist_to_text([n.nodeargd.argnlist[-1]]).strip().replace('   ', '&')
    if key:
        return '<cit page="{}">{}</cit>'.format(page.replace('"', ''), key)
    return ''


def langinfo(n, l2tobj):
    if not n.nodeargd:
        return ''
    res = ''
    for i, arg in enumerate(n.nodeargd.argnlist):
        t = l2tobj.nodelist_to_text([arg]).strip()
        if t:
            if i != 2:
                res += ' ' + t
            else:
                res += '<cit page="">{}</cit>'.format(t)
    return res


macros = [
    latex2text.MacroTextSpec('section',
     lambda n, l2tobj: u'\n\n{}\n'.format(l2tobj.node_arg_to_text(n, 2))),
    latex2text.MacroTextSpec("footnote", simplify_repl=footnote),
    latex2text.MacroTextSpec("footnotetext", simplify_repl=footnote),
    latex2text.MacroTextSpec("japhdoi", simplify_repl=japhdoi),
    latex2text.MacroTextSpec("japhug", simplify_repl=japhug),
    latex2text.MacroTextSpec("textup", simplify_repl=firstarg),
    latex2text.MacroTextSpec("textupsc", simplify_repl=uppercase_arg),
    latex2text.MacroTextSpec("glossfeat", simplify_repl=uppercase_arg),
    latex2text.MacroTextSpec("linieb", simplify_repl=secondarg),
    latex2text.MacroTextSpec("langinfo", simplify_repl=langinfo),
    latex2text.MacroTextSpec("ulp", simplify_repl=firstarg),
    latex2text.MacroTextSpec("ulg", simplify_repl=firstarg),
    latex2text.MacroTextSpec("ref", simplify_repl=firstarg),
    latex2text.MacroTextSpec("textsc", simplify_repl=uppercase_arg),
    latex2text.MacroTextSpec("Sc", simplify_repl=uppercase_arg),
    latex2text.MacroTextSpec("tsc", simplify_repl=uppercase_arg),
    latex2text.MacroTextSpec("tss", simplify_repl=dot_uppercase_arg),
    latex2text.MacroTextSpec("gsc", simplify_repl=uppercase_arg),
    latex2text.MacroTextSpec("ig", simplify_repl=uppercase_arg),
    latex2text.MacroTextSpec("mc", simplify_repl=uppercase_arg),
    latex2text.MacroTextSpec("gloss", simplify_repl=uppercase_arg),
    latex2text.MacroTextSpec("llap", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("ilt", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("il", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("is", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("ist", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("ia", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("ix", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("ux", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("dline", simplify_repl=lambda *args: ''),
    # latex2text.MacroTextSpec("ili", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("ili", simplify_repl=firstarg),
    latex2text.MacroTextSpec("REF", simplify_repl=lambda *args: ''),
    latex2text.MacroTextSpec("particle", simplify_repl='PARTICLE'),
    # For example parsing, we want to disregard jambox content
    # latex2text.MacroTextSpec("jambox", simplify_repl=''),
    # But for citation parsing, we need it.
    latex2text.MacroTextSpec("jambox", simplify_repl=lastarg),
    latex2text.MacroTextSpec("mbox", simplify_repl=firstarg),
    latex2text.MacroTextSpec("blockcquote", simplify_repl=fatcit),
    latex2text.MacroTextSpec("scite", simplify_repl=scite),
    latex2text.MacroTextSpec("fatcit", simplify_repl=fatcit),
    latex2text.MacroTextSpec("fatcitNP", simplify_repl=fatcit),
    latex2text.MacroTextSpec("textcquote", simplify_repl=cite),
    latex2text.MacroTextSpec("pgcitet", simplify_repl=cite),
    latex2text.MacroTextSpec("posscite", simplify_repl=cite),
    latex2text.MacroTextSpec("posscitet", simplify_repl=cite),
    latex2text.MacroTextSpec("posscitealt", simplify_repl=cite),
    latex2text.MacroTextSpec("namecite", simplify_repl=cite),
    latex2text.MacroTextSpec("Textcite", simplify_repl=cite),
    latex2text.MacroTextSpec("textcite", simplify_repl=cite),
    latex2text.MacroTextSpec("textcites", simplify_repl=cite),
    latex2text.MacroTextSpec("nocite", simplify_repl=cite),
    latex2text.MacroTextSpec("parencite", simplify_repl=cite),
    latex2text.MacroTextSpec("href", simplify_repl=href),
    latex2text.MacroTextSpec("possessivecite", simplify_repl=cite),
    latex2text.MacroTextSpec("autocite", simplify_repl=cite),
    latex2text.MacroTextSpec("autocites", simplify_repl=cite),
    latex2text.MacroTextSpec("cite", simplify_repl=cite),
]
for k, v in SIMPLE_MACROS.items():
    macros.append(latex2text.MacroTextSpec(k, simplify_repl=functools.partial(repl, v)),)
for abbr in lgr.ABBRS:
    if abbr:
        macros.extend([
            latex2text.MacroTextSpec(abbr, simplify_repl=functools.partial(repl, abbr)),
            latex2text.MacroTextSpec(abbr.lower(), simplify_repl=functools.partial(repl, abbr)),
        ])
        if len(abbr) > 1:
            macros.append(
                latex2text.MacroTextSpec(
                    abbr.capitalize(), simplify_repl=functools.partial(repl, abbr)))

l2t_context_db = latex2text.get_default_latex_context_db()
l2t_context_db.add_context_category('gll', prepend=True, macros=macros[:])

simple_converter_context_db = latex2text.get_default_latex_context_db()
simple_converter_context_db.add_context_category(
    'simple',
    prepend=True,
    macros=[latex2text.MacroTextSpec("href", simplify_repl=firstarg)])


def custom_latex_to_text(input_latex, parser=lw_context_db, converter=l2t_context_db):
    # the latex parser instance with custom latex_context
    lw_obj = latexwalker.LatexWalker(input_latex, latex_context=parser)
    # parse to node list
    nodelist, pos, length = lw_obj.get_latex_nodes()
    # initialize the converter to text with custom latex_context
    l2t_obj = latex2text.LatexNodes2Text(latex_context=converter)
    # convert to text
    try:
        return l2t_obj.nodelist_to_text(nodelist)
    except (IndexError, ValueError):
        return input_latex


def simple_to_text(latex):
    return custom_latex_to_text(
        latex, parser=simple_parser_context_db, converter=simple_converter_context_db)


def strip_tex_comment(s):
    lines = re.split(r'\\\\', s)
    if len(lines) == 2 and lines[1].startswith('%'):
        s = lines[0] + r'\\'
    return re.split(r"(?<!\\)%", s)[0].replace(r"\%", "%")


def iter_abbreviations(tex):
    for line in tex.split('\n'):
        if not line.strip().startswith("%"):
            cells = line.split("&")
            if len(cells) == 2:
                yield to_text(cells[0].strip())[0], to_text(cells[1].strip())[0]


def to_text(latex):
    # preprocessing:
    latex = latex.replace(r'{\sc ', r'\textsc{')
    latex = latex.replace(r'{\scshape ', r'\textsc{')

    # custom latex-to-text conversion:
    text, comment, refs = custom_latex_to_text(latex), [], []

    # postprocessing:
    text = text.replace('<backslash>', '\\')
    text = text.strip()
    # extract footnotes:
    fn_pattern = re.compile(r'§fn§([^§]+)§/fn§')
    for m in fn_pattern.finditer(text):
        comment.append(m.groups()[0])
    if comment:
        text = fn_pattern.sub('', text).strip()

    # extract citations:
    pattern = re.compile(r'<cit page="([^"]*)">([^<]+)</cit>')
    for m in pattern.finditer(text):
        if m.groups()[1] != '[':
            for sid in m.groups()[1].split(','):
                if sid.strip():
                    refs.append((sid.strip(), m.groups()[0]))
    if refs:
        text = pattern.sub('', text).strip()

    for cc in comment:
        for m in pattern.finditer(cc):
            if m.groups()[1] != '[':
                for sid in m.groups()[1].split(','):
                    if sid.strip():
                        refs.append((sid.strip(), m.groups()[0]))
    comment = [pattern.sub(lambda m: m.groups()[1], cc).strip() for cc in comment]

    #
    # FIXME: handle *\textit commands
    #
    leading_amp = re.compile(r'^\s*&+\s*')
    #
    # FIXME: unicodedata.normalize.nfc!
    #
    return (
        text,
        '\n'.join(comment),
        [(leading_amp.sub('', k.replace('–', '--').lower()), v) for k, v in refs])



"""
Functionality to merge bibliographies serialized as BibTeX files.
"""
import re
import typing
import pathlib
import collections

from clldutils.misc import slug
from pybtex import database
from pycldf.sources import Source
from thefuzz import fuzz
from unidecode import unidecode
from pylatexenc.latex2text import LatexNodes2Text

__all__ = ['iter_merged', 'iter_entries', 'merge']

YEAR_PATTERN = re.compile('([0-9]{4})')
ACC_FIELDS = {  # Fields where content from merged records should be accumulated.
    'isreferencedby': ' ',
}


def merge(d: pathlib.Path, bib: pathlib.Path, delatex=False):
    res = []
    for src, _ in iter_merged(iter_entries(d, delatex=delatex)):
        res.append(src.bibtex())
    bib.write_text('\n'.join(res), encoding='utf8')
    return len(res)


def hash(e):
    creators = e.persons.get('author') or e.persons.get('editor')
    year = YEAR_PATTERN.search(e.fields.get('year') or '')
    return (
        slug(e.fields.get('title') or ''),
        year.groups()[0] if year else '',
        slug(''.join(creators[0].last_names)) if creators else ''
    )


def similarity(s1, s2):
    return max([fuzz.ratio(s1, s2), fuzz.token_set_ratio(s1, s2)])


def iter_entries(
        d: typing.Union[str, pathlib.Path],
        delatex=False) -> typing.Generator[database.Entry, None, None]:
    d = pathlib.Path(d)
    if d.is_dir():
        paths = sorted(d.glob('*.bib'), key=lambda pp: pp.stem)
    else:
        paths = [d]
    for p in paths:
        for e in database.parse_string(p.read_text(encoding='utf8'), 'bibtex').entries.values():
            if delatex:
                src = Source.from_entry(e.key, e)
                for k in list(src.keys()):
                    try:
                        src[k] = LatexNodes2Text().latex_to_text(src[k])
                    except:  # noqa: E722
                        pass
                e = src.entry
                e.key = src.id
            yield e


def iter_merged(
        entries: typing.Iterable[typing.Union[Source, database.Entry]]
) -> typing.Generator[typing.Tuple[Source, dict], None, None]:
    """
    We merge in a multi-step procedure.

    1. We compute a hash based on creators, year and title for each entry, and aggregate entries
       with matching hash.
    2. We compute a citation key based on creators and year, and if entries with different hashes
       get the same key, we examine the titles using fuzzy comparison.
    3. The final citation key is created from the provisional keys by appending a number for
       disambiguation.
    """
    aggr = collections.defaultdict(list)
    for e in entries:
        if isinstance(e, Source):
            sid = e.id
            e = e.entry
            e.key = sid
        aggr[hash(e)].append(e)

    keys = {}
    by_key = collections.defaultdict(list)

    def norm_title(e):
        return e.fields.get('title', '').lower()

    for h, v in sorted(aggr.items()):
        t, y, a = h
        if t and (y or a):  # A meaningful has. So we assume all entries in v to be identical.
            k = make_key(v[0])
            if k in keys:  # The same key has already been computed for a batch of entries.
                (_, y2, a2), title2 = keys[k]
                title = norm_title(v[0])

                sim = 0
                if y == y2 and a == a2 and title and title2:
                    sim = similarity(title, title2)

                if sim >= 92:  # Just a threshold ...
                    # identify with the previous batch under this key!
                    by_key[k][-1].extend(v)
                else:
                    # Otherwise, just add a new batch for the key:
                    by_key[k].append(v)
            else:
                by_key[k].append(v)
            keys[k] = h, norm_title(v[0])
        else:  # not enough information to actually identify records!
            for e in v:
                k = make_key(e)
                by_key[k].append([e])

    for k, batches in sorted(by_key.items(), key=lambda i: i[0]):
        if len(batches) == 1:
            yield merged(k, batches[0])
        else:
            for i, batch in enumerate(batches, start=1):
                yield merged('{}:{}'.format(k, i), batch)


def merged(key, batch):
    m, keymap = None, {}
    for i, e in enumerate(batch):
        keymap[e.key] = key
        if i == 0:
            m = Source.from_entry(key, e)
        else:
            for k, v in e.fields.items():
                if not v.strip():  # pragma: no cover
                    continue
                if k in ACC_FIELDS:
                    if k in m:
                        m[k] += '{}{}'.format(ACC_FIELDS[k], v)
                    else:
                        m[k] = v
                else:
                    if k not in m or len(m[k]) < len(v):  # longest is best
                        m[k] = v
    m['citekeys'] = ' '.join(sorted(keymap))
    return m, keymap


def make_key(e):
    ed = ''
    creators = e.persons.get('author')
    if not creators:
        creators = e.persons.get('editor')
        if creators:
            ed = ':ed'
    if creators:
        s = ''
        for i, p in enumerate(creators, start=1):
            if i > 2:
                s += ':etal'
                break
            s += (':' if s else '') + ''.join(p.prelast_names + p.last_names)
    else:
        s = 'np'
    year = YEAR_PATTERN.search(e.fields.get('year') or '')
    if year:
        year = year.groups()[0][2:]
    else:
        year = 'nd'
    s = s.replace("ä", "ae")
    s = s.replace("ö", "oe")
    s = s.replace("ü", "ue")
    s = s.replace('"=', '-')
    creators = unidecode(s)
    for c in "/.'()= ":
        creators = creators.replace(c, '')
    return creators.lower() + ed + ':' + year



"""
Show the IGT examples of a publication
"""
from linglit.cli_util import add_publication, get_publication


def register(parser):
    add_publication(parser)


def run(args):
    for ex in get_publication(args).examples:
        print(ex)
        print('')



"""
Functionality to read LSP BibTeX files.
"""
import re
import typing
import logging
import pathlib
import subprocess
import unicodedata

from pybtex import database
from clldutils.source import Source
from clldutils.misc import slug
from clldutils.path import ensure_cmd

from .latex import simple_to_text
from . import cfg

__all__ = ['iter_bib', 'normalize_key']

# Invalid author lists (and fixes):
NAMES = {
    'SkL, Holm. B 74, 4\\textsuperscript{o} = Brøndum-Nielsen, Johs':
        'SkL, Holm. B and Brøndum-Nielsen, Johs',
    'De Marneffe, Marie-Catherine, Timothy Dozat, Natalia Silveira, Katri Haverinen, Filip Ginter, '
    'Joakim Nivre,':
        'De Marneffe, Marie-Catherine and Timothy Dozat and Natalia Silveira and Katri Haverinen '
        'and Filip Ginter and Joakim Nivre,',
    'Igoo Ribaa, Mark Post,Ilii Ribaa, Miilii Nodu, Kenjum Bagra, Bomcak Ribaa, Toomoo Ribaa, '
    'Notoo Aado, Dambom Keenaa':
        'Igoo Ribaa and Mark Post and Ilii Ribaa and Miilii Nodu and Kenjum Bagra and Bomcak Ribaa '
        'and Toomoo Ribaa and Notoo Aado and Dambom Keenaa',
    'Ringe, Donald A., Jr.,':
        'Ringe, Donald A. Jr.,',
    'Yahalom-Mack, Naama, Eliyahu-Behar, Adi':
        'Yahalom-Mack, Naama and Eliyahu-Behar, Adi',
    'Vajrabhaya, P., Kapatsinski, Vsevolod':
        'Vajrabhaya, P. and Kapatsinski, Vsevolod',
    'Kroch, Anthony, Santorini, Beatrice':
        'Kroch, Anthony and Santorini, Beatrice',
    'Féry, Caroline, Fanselow, Gisbert':
        'Féry, Caroline and Fanselow, Gisbert',
    'Dan Velleman, David Beaver, Emilie Destruel, Dylan Bumford, Edgar Onea, Liz Coppock':
        'Dan Velleman and David Beaver and Emilie Destruel and Dylan Bumford and Edgar Onea and '
        'Liz Coppock',
    'Riad, Tomas, & Gussonhoven, Carlos':
        'Riad, Tomas and Gussonhoven, Carlos',
    "B\\'{e}r\\'{e}nice Bellina, Elisabeth A Bacus, Thomas Oliver Pryce, Jan Wisseman Christie":
        "B\\'{e}r\\'{e}nice Bellina and Elisabeth A Bacus and Thomas Oliver Pryce and Jan "
        "Wisseman Christie",
    "D'Alessandro, Roberta \\& Fischer, Susann \\& Hrafnbjargarson, Gunnar Hrafn":
        "D'Alessandro, Roberta and Fischer, Susann and Hrafnbjargarson, Gunnar Hrafn",
    'Mitchell, Rosamond, Tracy-Ventura, Nicole':
        'Mitchell, Rosamond and Tracy-Ventura, Nicole',
    'Juan-Garau, Maria., Joana Salazar-Noguera,':
        'Juan-Garau, Maria and Joana Salazar-Noguera',
    'Andr{\\\'e} M{\\"u}ller, Annkathrin Wett, Viveka Velupillai, Julia Bischoffberger, '
    'Cecil H. Brown, Eric W. Holman, Sebastian Sauppe, Zarina Molochieva, Pamela Brown, '
    'Harald Hammarstr{\\"o}m, Oleg Belyaev, Johann-Mattis List, Dik Bakker, Dmitry Egorov, '
    'Matthias Urban, Robert Mailhammer, Agustina Carrizo, Matthew S. Dryer, Evgenia Korovina, '
    'David Beck, Helen Geyer, Patience Epps, Anthony Grant,':
        'Andr{\\\'e} M{\\"u}ller and Annkathrin Wett and Viveka Velupillai and Julia '
        'Bischoffberger and Cecil H. Brown and Eric W. Holman and Sebastian Sauppe and Zarina '
        'Molochieva and Pamela Brown and Harald Hammarstr{\\"o}m and Oleg Belyaev and '
        'Johann-Mattis List and Dik Bakker and Dmitry Egorov and Matthias Urban and Robert '
        'Mailhammer and Agustina Carrizo and Matthew S. Dryer and Evgenia Korovina and David Beck '
        'and Helen Geyer and Patience Epps and Anthony Grant',
}

GENRE_MAP = {
    "report": "techreport",
    "thesis": "phdthesis",
    "webpage": "online",
    "mphillthesis": "mastersthesis",
    "unpublished": "misc",
    "unpbulished": "misc",
    "mvbook": "book",
    "url": "online",
    "unpub": "misc",
    "data": "misc",
    "other": "misc",
    "customa": "misc",
    "mathesis": "mastersthesis",
    "hmisc": "misc",
}

FIELD_MAP = {
    "journaltitle": "journal",
    "ids": "ids",
    "urldate": "",
    "isbn": "isbn",
    "subtitle": "subtitle",
    "edition": "edition",
    "type": "type",
    "institution": "institution",
    "shorttitle": "",
    "month": "",
    "howpublished": "howpublished",
    "issn": "issn",
    "abstract": "",
    "booksubtitle": "booksubtitle",
    "issue": "issue",
    "organization": "organization",
    "venue": "venue",
    "langid": "",
    "annote": "",
    "annotation": "",
    "language": "",
    "numpages": "pages",
    "stableurl": "url",
    "addendum": "",
    "shortauthor": "",
    "chapter": "chapter",
    "file": "",
    "shortjournal": "journal",
    "optaddress": "",
    "call-number": "",
    "sortyear": "",
    "lgcode": "lgcode",
    "options": "",
    "namea": "",
    "maintitle": "",
    "acmid": "",
    "pubstate": "",
    "opturl": "url",
    "translator": "",
    "eprint": "eprint",
    "related": "",
    "alnumcodes": "",
    "languoidbase_ids": "",
    "numberofpages": "pages",
    "day": "",
    "bib": "",
    "eventtitle": "eventtitle",
    "optmonth": "",
    "collaborator": "",
    "optchapter": "",
    "titleaddon": "",
    "sortkey": "",
    "serie": "series",
    "optnote": "",
    "editors": "editor",
    "relatedstring": "",
    "lastchecked": "",
    "issuetitle": "",
    "optannote": "",
    "bdsk-url-2": "",
    "autheditor": "editor",
    "biburl": "",
    "opturldate": "",
    "bdsk-file-2": "",
    "read": "",
    "key": "key",
    "bibsource": "",
    "optkey": "",
    "pagetotal": "pages",
    "lccn": "lccn",
    "origdate": "",
    "sortauthor": "",
    "optnumber": "",
    "place": "address",
    "issue_date": "",
    "hyphenation": "",
    "urlaccessdate": "",
    "adress": "address",
    "archiveprefix": "",
    "version": "version",
    "eid": "",
    "optlocation": "",
    "eprinttype": "",
    "page": "pages",
    "lid": "",
    "chaper": "chapter",
    "author-translation": "",
    "rating": "",
    "creationdate": "",
    "opttype": "",
    "bdsk-file-3": "",
    "eventdate": "",
    "price": "",
    "booktitleaddon": "",
    "collection": "",
    "mendeley-groups": "",
    "urlmonth": "",
    "urlday": "",
    "vol": "volume",
    "optseries": "series",
    "inlg": "inlg",
    "macro_area": "",
    "pmid": "",
    "interhash": "",
    "intrahash": "",
    "oldhhfn": "",
    "src": "",
    "pdf": "",
    "editora": "editor",
    "editoratype": "",
    "sorttitle": "",
    "hhtype": "hhtype",
    "city": "address",
    "added-at": "",
    "optaddendum": "",
    "ozbib_id": "",
    "volumes": "",
    "hal_id": "",
    "hal_version": "",
    "bookauthor": "",
    "xjournal": "",
    "xpublisher": "",
    "xeditor": "",
    "opteditor": "",
    "ee": "",
    "ozbibreftype": "",
    "pubkey": "",
    "originalyear": "",
    "optpublisher": "",
    "journalsubtitle": "",
    "citeulike-article-id": "",
    "posted-at": "",
    "shorthand": "",
    "numer": "number",
    "addres": "address",
    "optxref": "",
    "optvolume": "",
    "groups": "",
    "fn": "",
    "annotate": "",
    "aiatsis_callnumber": "",
    "oclcid": "",
    "url_checked": "",
    "homepage": "",
    "uselessnote": "",
    "more-editors": "",
    "publishers": "publisher",
    "citedin": "",
    "optcrossref": "",
    "optedition": "",
    "primaryclass": "",
    "rights": "rights",
    "yaddress": "",
    "booktitel": "booktitle",
    "typ": "",
    "iso_code": "lgcode",
    "olac_field": "olac_field",
    "wals_code": "wals_code",
    "oclc": "oclc",
    "editorx": "",
    "xisbn": "",
    "aiatsis_code": "aiatsis_code",
    "aiatsis_reference_language": "",
    "accessdate": "",
    "institutions": "institution",
    "issue_year": "",
    "book": "",
    "research_area": "",
    "urlpage": "",
    "optpages": "",
    "quality": "",
    "author-transl": "",
    "origlanguage": "",
    "bootitle": "booktitle",
    "bdsk-file-4": "",
    "vols": "",
    "optbooktitle": "",
    "altauthor": "",
    "optorganization": "",
    "email": "",
    "issuesubtitle": "",
    "optvolumes": "",
    "cfn": "",
    "class_loc": "",
    "document_type": "",
    "mpi_eva_library_shelf": "",
    "optauthor": "",
    "notes": "",
    "ozbibnote": "",
    "asjp_name": "",
    "indexauthor": "",
    "urlx": "",
    "reprint": "",
    "optsubtitle": "",
    "opttitle": "",
    "mainsubtitle": "",
    "issueeditor": "",
    "anote": "",
    "articleno": "",
    "optschool": "",
    "research_sub_area": "",
    "urldsate": "",
    "urlzip": "",
    "origyear": "",
    "postscript": "",
    "editorbtype": "",
    "nourl": "",
    "optshorthand": "",
    "execute": "",
    "urladte": "",
    "eventvenue": "",
    "optrelated": "",
    "optrelatedtype": "",
    "seriestitle": "",
    "optassress": "",
    "db": "",
    "tags": "",
    "affiliation": "",
    "ean": "",
    "status": "",
    "eidtor": "editor",
    "publihers": "publisher",
    "risfield_0_bt": "",
    "paper": "",
    "class": "",
    "double_booktitle": "",
    "numnber": "number",
    "numbers": "number",
    "publihser": "publisher",
    "mpifn": "",
    "subject_headings": "",
    "announced": "",
    "paperbackoptisbn": "",
    "pgaes": "pages",
    "added": "",
    "refdb_id": "",
    "booktitle1": "",
    "addresss": "address",
    "annotex": "",
    "issnx": "",
    "publisherx": "",
    "venuedate": "",
    "editior": "editor",
    "priority": "",
    "part": "part",
    "dddress": "address",
    "opttitleaddon": "",
    "local-noopurl": "",
    "annnote": "",
    "publishing": "",
    "utf_author": "",
    "country": "",
    "subject": "",
    "publication": "",
    "institute": "school",
    "sannote": "",
    "tppubtype": "",
    "scshool": "school",
    "preis": "",
    "opthowpublished": "",
    "lastcheck": "",
    "bdsk-file-5": "",
    "bdsk-file-6": "",
    "bdsk-file-7": "",
    "journa": "journal",
    "source": "",
    "unmatched-authority": "",
    "dewey-call-number": "",
    "library-id": "",
    "size": "",
    "etc": "",
    "adderss": "address",
    "addrss": "address",
    "origtitle": "",
    "origlocation": "",
    "local-url": "",
    "misc": "",
    "longnote": "",
    "uuid": "",
    "introduction": "",
    "entrysubtype": "",
    "optissn": "",
    "verkauf": "",
    "optyear": "",
    "adsnote": "",
    "adsurl": "",
    "titel": "title",
    "optlanguage": "",
    "publsher": "publisher",
    "refid": "",
    "websitetitle": "",
    "booktile": "booktitle",
    "arxivid": "",
    "description": "",
    "ris_reference_number": "",
    "citeulike-linkout-0": "",
    "eventdatet": "",
    "bdsk-url-3": "",
    "bdsk-url-4": "",
    "sssauthor": "",
    "sssbooktitle": "",
    "optstableurl": "",
    "itemartakarmat": "",
    "yomi": "",
    "substitle": "subtitle",
    "bdsk-url": "",
    "types": "",
    "noter": "",
    "orig_": "",
    "title_english": "",
    "alteditor": "",
    "bookstitle": "booktitle",
    "unidentified": "",
    "sorname": "",
    "scool": "school",
    "bootktitle": "booktitle",
    "note_original": "",
    "abbrv": "",
    "confdate": "",
    "xdoi": "",
    "xurl": "",
    "editorb": "",
    "label": "",
    "vollume": "volume",
    "trad": "",
    "library": "",
    "elocation-id": "",
    "pooktitle": "booktitle",
    "attachments": "",
    "shorteditor": "",
    "xxxauthor": "",
    "pubisher": "publisher",
    "translation": "",
    "fnnote": "",
    "autor": "author",
    'modified': '',
    'authauthor': "author",
    'bdsk-url-1': "",
    'optisbn': "",
    'cited': "",
    'comment': "",
    'bilal': "",
    'achilleos': "",
    'xmonth': "",
    'xaddress': "",
    'last_changed': "",
    'urlyear': "",
    'copyright': "",
}


def normalize_key(k):
    return unicodedata.normalize('NFC', k.replace('–', '--').lower())


class LangsciSource(Source):
    """
    Some bibfiles provide fields "IDs" or "ids" which list alternative keys by which a record can
    be cited in the tex files.
    """
    @property
    def alt_keys(self):
        return set([normalize_key(s.strip()) for s in self.get('ids', '').split(',') if s.strip()])


def to_source(key, e):
    def remove_unbalanced_braces(s):
        for o, c in ['()', '{}']:
            if s.count(o) != s.count(c):
                s = s.replace(o, '').replace(c, '')
        return s

    src = LangsciSource.from_entry('x', e) if not isinstance(e, LangsciSource) else e
    src.id = normalize_key(key)
    src.genre = GENRE_MAP.get(src.genre.lower(), src.genre.lower())
    if src.genre == 'phdthesis' and (
            slug(src.get('type', '')) == 'mathesis' or 'M.A.' in src.get('type', '')):
        src.genre = 'mastersthesis'
        del src['type']
    for k in list(src):
        src[k.lower() if k.lower() != 'date' else 'year'] = remove_unbalanced_braces(
            simple_to_text(src.pop(k)))
        if k in ['author', 'editor']:
            src[k] = src[k].replace('[', '').replace(']', '')

    for field, value in list(src.items()):
        if field in FIELD_MAP:
            target = FIELD_MAP[field]
            if target:
                if target not in src:
                    src[target] = value
                if target != field:  # A field is renamed.
                    del src[field]
            else:
                del src[field]

    return src


def iter_bib(ps: typing.List[pathlib.Path], verbose=False) -> typing.Generator[Source, None, None]:
    """
    :param ps: `list` of paths to bibtex files making up one bibliography.
    """
    log = logging.getLogger(__name__)
    bibtex = []
    for p in ps:
        # preprocess the bibtex, fixing the stuff that bibtool can't fix:
        text = p.read_text(encoding='utf8')
        lines = [ln.strip() for ln in text.split('\n') if ln.strip()]
        if len(lines) == 1 and len(lines[0]) < 200 and p.parent.joinpath(lines[0]).exists():
            # Special handling for 237, where the path to 223's bib is given in the bibfile!
            text = p.parent.joinpath(lines[0]).read_text(encoding='utf8')  # pragma: no cover

        text = text.replace('\xa0', ' ')

        for pattern, repl in [
            # Invalid field names:
            (r'^\s*\_orig\s*=', '  orig ='),
            (r'^\s*\@(Note|Abstract|Comments?)\s*=', '  xnote ='),
            (r'^\s*\_\_(M|m)(arkedentry|ARKEDENTRY)\s*=', '  markedentry ='),
            (r'^\s*\_\_publisher\s*=', '  publisher ='),
            (r'^\s*\-\-Address\s*=', '  xxaddress ='),
            (r'^\s*\-\-Author\s*=', '  xxauthor ='),
            # Commented lines:
            (r'^\s*%.*$', ''),
            # Trailing comments:
            (r',%([0-9a-z]*|Undefined\?|(N|n)ame is not available)$', ','),
            # Invalid keys:
            (r'@([a-z]+)\{\{(English|French|vedic)\}',
             lambda m: '@%s{%s_' % (m.groups()[0], m.groups()[1])),
            # crossref to id instead of key:
            (r'crossref\s*=\s*{dryhaspWALS}', 'crossref = {wals}'),
            (r'crossref\s*=\s*{coling98}', 'crossref = {Branco98a}'),
        ]:
            text = re.sub(pattern, repl, text, flags=re.MULTILINE)

        # Single instances of weirdness:
        for k, v in [
            (r'{{\{', '{{{'),
            (r'{L\{', '{L{\\'),
            ('}%,', '},'),
            (r'{\{AA}}', r'{\AA}'),
            ('RepúblicadelParaguay2001,', 'RepublicadelParaguay2001,'),
            (r'@book{Kury\l{}owicz1973', '@book{kurylowicz1973'),
            ('@Article{,', '@Article{undefined,'),
            ('@book:1997\n{foley,', '@book{foley:1997,'),
        ]:
            text = text.replace(k, v)
        bibtex.append(text)

    # Now run bibtool:
    cmd = subprocess.Popen(
        [ensure_cmd('bibtool'), '-r', str(cfg.BIBTOOL_RSC)],
        stdin=subprocess.PIPE,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    stdout, stderr = cmd.communicate(input='\n'.join(bibtex).encode('utf8'))
    for line in stderr.decode('utf8').splitlines():  # pragma: no cover
        if line.strip():
            if any(s in line for s in
                   ['Duplicate field', 'non-space characters ignored', "Missing ',' assumed"]):
                continue
            if verbose:
                log.warning('bibtool:::' + line)

    # Now feed the bibtex to pybtex one record at a time to skip duplicate keys:
    for i, chunk in enumerate(re.split(r'^@', stdout.decode('utf8'), flags=re.MULTILINE)):
        # Again some preprocessing:
        if re.match('[A-Za-z]+{,', chunk):  # record without key.
            continue  # pragma: no cover
        if chunk.startswith('unpublished{rien'):  # empty stub.
            continue  # pragma: no cover
        for k, v in NAMES.items():  # Fix invalid author lists.
            chunk = chunk.replace(k, v)
        # Field values without enclosing braces:
        chunk = re.sub(
            r'^\s*([a-zA-Z]+)\s*=([^{]+),$',
            lambda m: '%s = {%s},' % (m.groups()[0], m.groups()[1]),
            chunk,
            flags=re.MULTILINE)
        if i:
            try:
                for k, e in database.parse_string('@' + chunk, 'bibtex').entries.items():
                    yield to_source(k, e)
            except:  # pragma: no cover # noqa: E722
                raise ValueError('{}::@{}'.format(ps, chunk))  # pragma: no cover



import re

from csvw.dsv import reader
from pyglottolog.languoids import Glottocode

from linglit import util

CFG_PATH = util.CFG_PATH / 'glossa'
LNAME_MAP = {
    'Medieval Spanish': 'Old Spanish',
}


def language_specs():
    res = {}
    for d in reader(CFG_PATH / 'glossa.csv', dicts=True):
        if d['example_languages']:
            res[int(d['id'])] = LanguageSpec(d['example_languages'])
    return res


class LanguageSpec:
    def __init__(self, spec):
        self.language_ranges = {}
        self.language = None
        if spec.strip():
            chunks = [s.strip() for s in spec.split(',')]
            if len(chunks) == 1 and Glottocode.pattern.fullmatch(chunks[0]):
                self.language = chunks[0]
            elif chunks:
                for chunk in chunks:
                    glottocode, _, range = chunk.partition(':')
                    glottocode = glottocode.strip()
                    assert Glottocode.pattern.fullmatch(glottocode)
                    lower, _, upper = range.partition('-')
                    lower = int(lower) if lower.strip() else 0
                    upper = int(upper) if upper.strip() else 1000
                    self.language_ranges[glottocode] = (lower, upper)

    def __call__(self, name, lid):
        # 1. Check, if ex has already a resolvable Language_Name
        if name:
            return name
        if self.language:
            return self.language
        if self.language_ranges:
            # Get the numeric part of the local ID
            m = re.match(r'[0-9]+', lid)
            if m:
                lid = int(lid[:m.end()])
                for gc, (min, max) in self.language_ranges.items():
                    if min <= lid <= max:
                        return gc



"""
Show the bibliography of a publication
"""
from linglit.cli_util import add_publication, get_publication


def register(parser):
    add_publication(parser)
    parser.add_argument('--bibtex', action='store_true', default=False)


def run(args):
    for src in get_publication(args).cited_references:
        print(src.bibtex() if args.bibtex else src)
        print('')



import attr
from pycldf.sources import Source

from linglit import base
from . import xml


@attr.s
class Record(base.Record):
    doc = attr.ib()

    def as_source(self):
        return Source(
            'article',
            '{}'.format(self.ID),
            author=self.creators,
            title=self.title,
            year=self.year,
            volume=self.doc.xpath(".//volume")[0].text,
            issue=self.doc.xpath(".//issue")[0].text,
            doi=self.DOI,
            publisher="Open Library of Humanities",
            journal="Glossa: a journal of general linguistics",
        )

    @property
    def has_open_license(self):
        return True


class Publication(base.Publication):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        self.doc = xml.parse(self.dir)
        self.language_spec = self.record
        self.record = Record(**xml.metadata(self.dir, self.doc))
        self.abbreviations = xml.abbreviations(self.doc)

    def iter_references(self):
        yield from xml.refs(self.doc)

    def iter_cited(self):
        for xref in self.doc.xpath(".//xref[@ref-type='bibr']"):
            yield xref.get('rid')

    def iter_examples(self, glottolog=None):
        for count, number, letter, lang, xrefs, igt, comment in xml.iter_igt(
                self.doc, self.abbreviations):
            lid = '{}{}'.format(number or '', letter or '')
            refs = []
            for sid, reft, label in xrefs:
                if reft == 'bibr':
                    pages = label.partition(':')[2].strip() if label else ''
                    refs.append((sid, pages.replace('[', '(').replace(']', ')')))

            yield base.Example(
                ID='{}'.format(count),
                Local_ID=lid,
                Primary_Text=igt.primary_text,
                Analyzed_Word=igt.phrase,
                Gloss=igt.gloss,
                Translated_Text=igt.translation,
                Language_ID=None,
                Language_Name=self.language_spec(lang, lid) if self.language_spec else lang,
                Source=refs,
                Abbreviations=self.abbreviations,
                Meta_Language_ID='stan1293',
                Comment=comment,
            )



import re
import urllib.request

from bs4 import BeautifulSoup as bs

from linglit import base
from .publication import Publication
from . import cfg

BASE_URL = "https://www.glossa-journal.org"
URL_PATTERN = re.compile('article/id/(?P<id>[0-9]+)')
CATALOG_URL = BASE_URL + "/articles/"


class Repository(base.Repository):
    id = 'glossa'
    lname_map = cfg.LNAME_MAP

    def create(self, verbose=False):
        get_all(self.dir, verbose=verbose, pages=30)

    def __getitem__(self, item):
        p = self.dir / '{}.xml'.format(item)
        if not p.exists():
            raise KeyError(item)
        lspecs = cfg.language_specs()
        return Publication(lspecs.get(int(item)), p, repos=self)

    def iter_publications(self):
        lspecs = cfg.language_specs()
        for p in self.dir.glob('*.xml'):
            yield Publication(lspecs.get(int(p.stem)), p, repos=self)


def download(url, verbose=False):
    if verbose:
        print('retrieving {}'.format(url))
    return urllib.request.urlopen(url).read().decode('utf8')


def get_xml(url, d, verbose=False):
    if url:
        m = URL_PATTERN.search(url)
        if m:
            p = d / '{}.xml'.format(m.group('id'))
            if not p.exists():
                page = bs(download(url, verbose=verbose), 'lxml')
                xml_url = page.find('a', href=True, string='Download XML')
                if xml_url:
                    xml_url = xml_url['href']
                    p.write_text(download(BASE_URL + xml_url), encoding='utf8')
                else:  # pragma: no cover
                    return
            return p.read_text(encoding='utf8')


def get_all(d, verbose=False, pages=None):
    url = CATALOG_URL
    pagenum = 0
    while url:
        pagenum += 1
        if pages and pagenum >= pages:
            break  # pragma: no cover
        page = bs(download(url, verbose=verbose), 'lxml')
        for p in page.find_all('div', class_='card-panel'):
            for a in p.find_all('a'):
                get_xml(BASE_URL + a['href'], d, verbose=verbose)
                # download HTML, look for "Download XML" link:
                # <a href="/article/5809/galley/21790/download/">Download XML</a>
                # https://www.glossa-journal.org/article/5821/galley/21844/download/
                break

        pagination = page.find('ul', class_='pagination')
        active = None
        for p in pagination.find_all('li'):
            if active:  # pragma: no cover
                url = CATALOG_URL + p.find('a')['href']
                break
            if 'active' in p['class']:
                active = True
        else:
            break



import re
import collections

from clldutils.path import walk
from clldutils.misc import lazyproperty

from linglit import base
from .bibtex import iter_bib, normalize_key
from .texsoup import iter_texsoup_lines
from .latex import to_text, iter_abbreviations
from . import texfixes
from .examples import iter_gll, make_example
from . import cfg

MAKEFILE_NAME = 'Makefile'
MAIN_TEX_NAMES = ['main.tex', 'book.tex']
BACKMATTER_NAME = 'backmatter.tex'
ABBREVIATIONS_NAME = 'abbreviations.tex'
CHAPTERS_NAME = 'chapters'
MAIN_TEX_EXCEPTIONS = {  # Cases where the main TeX file can't be found:
    49: 'ind-steelsbook.tex',
}
NO_BIB = [274]
NO_INCLUDES = [163]
FIX_FILENAMES = {
    'parts/bilinguismo': 'parts/biblinguismo',
}


class Publication(base.Publication):
    def __init__(self, *args, **kw):
        super().__init__(*args, **kw)
        if self.record.int_id in MAIN_TEX_EXCEPTIONS:  # pragma: no cover
            main = self.dir / MAIN_TEX_EXCEPTIONS[self.record.int_id]
        else:
            main = self._find_main_tex()
        assert main, self.record.ID
        self.main = main

        self._bibs = None
        self._includes = None
        self._includes_tex = {}
        self._refs = []

    def iter_examples(self):
        texfile2language = {}
        for row in cfg.iter_texfile_titles():
            if row['Language']:
                texfile2language[(int(row['Book_ID']), row['Filename'])] = row['Language']
        seen = set()
        for p in self.includes:
            for linfo, gll, prevline in iter_gll(self.read_tex(p)):
                ex = make_example(self, linfo, gll, prevline)
                if ex and (ex.ID not in seen):
                    ex.Source_Path = p
                    if self.repos:
                        ex.Source_Path = p.relative_to(self.repos.dir)
                    refs = []
                    for key, pages in ex.Source:
                        key = normalize_key(key)
                        if key in self.bibkeys:
                            refs.append((self.bibkeys[key], pages))
                    ex.Source = refs
                    if ex.Language_Name is None and \
                            (self.record.int_id, p.name) in texfile2language:
                        ex.Language_Name = texfile2language[(self.record.int_id, p.name)]
                    if ex.Language_Name is None and self.record.objectlanguage:
                        ex.Language_Name = self.record.objectlanguage
                    if str(p) in self.gloss_abbreviations:
                        ex.Abbreviations = self.gloss_abbreviations[str(p)]
                    elif None in self.gloss_abbreviations:
                        ex.Abbreviations = self.gloss_abbreviations[None]
                    ex.Meta_Language_ID = self.record.metalanguage
                    seen.add(ex.ID)
                    yield ex

    def iter_cited(self):
        relevant = self.includes + [self.main]
        if self.main.parent.joinpath(BACKMATTER_NAME).exists():
            relevant.append(self.main.parent.joinpath(BACKMATTER_NAME))
        for p in relevant:
            for line in self.read_tex(p, with_input=(p != self.main)).split(r'\\'):
                _, _, refs = to_text(line)
                for ref, _ in refs:
                    key = normalize_key(ref)
                    if key in self.bibkeys:
                        yield self.bibkeys[key]

    def iter_references(self):
        if not self._refs:
            self._refs = list(iter_bib(self.bibs))
        yield from iter(self._refs)

    # --- langsci specifics
    def read_tex(self, p, with_input=True):
        if str(p) not in self._includes_tex:
            self._includes_tex[str(p)] = texfixes.read_tex(p, with_input=with_input)
        return self._includes_tex[str(p)]

    @lazyproperty
    def gloss_abbreviations(self):
        abbr_pattern = re.compile(r'\\(sub)?section\*?(\[[^]]+])?{Abbreviations(\s+[A-Za-z]+)*}')
        section_pattern = re.compile(r'\\(?:sub)?section')
        res = collections.defaultdict(collections.OrderedDict)

        for p in self.includes:
            tex = self.read_tex(p)
            m = abbr_pattern.search(tex)
            if m:
                for k, v in iter_abbreviations(section_pattern.split(tex[m.end():])[0]):
                    res[str(p)][k] = v

        for p in walk(self.dir):
            if p.name == ABBREVIATIONS_NAME:
                for k, v in iter_abbreviations(self.read_tex(p)):
                    res[None][k] = v

        return res

    def _get_includes_and_bibs(self):
        self._includes, self._bibs = includes_and_bib(
            self.dir,
            self.main,
            CHAPTERS_NAME if self.main.parent.joinpath(CHAPTERS_NAME).exists() else 'indexed',
            no_bib=self.record.int_id in NO_BIB,
        )

    @property
    def bibs(self):
        if self._bibs is None:
            self._get_includes_and_bibs()  # pragma: no cover
        return self._bibs

    @lazyproperty
    def bibkeys(self):
        res = {}
        for src in self.iter_references():
            res[src.id] = src.id
            for altkey in src.alt_keys:
                res[altkey] = src.id
        return res

    @property
    def includes(self):
        if self._includes is None:
            self._get_includes_and_bibs()
        return self._includes

    def _find_makefile(self):
        p = self.dir / MAKEFILE_NAME
        if p.exists():
            return p
        for p in walk(self.dir, mode='files'):
            if p.name == MAKEFILE_NAME:
                return p

    def _find_by_documentclass(self, d):
        for p in d.glob('*.tex'):
            for line in p.read_text(encoding='utf8').split('\n'):
                if all(w in line for w in [r'\documentclass', 'number']):
                    return p

    def _find_main_tex(self):
        make = self._find_makefile()
        if make:
            for line in make.read_text(encoding='utf8').splitlines():
                line = line.strip()
                if line.startswith('xelatex'):
                    tex = '{}.tex'.format(line.split()[-1])
                    if tex and make.parent.joinpath(tex).exists():
                        return make.parent / tex
            for line in make.read_text(encoding='utf8').splitlines():
                pdf = re.search(r'\s+([A-Za-z_0-9]+)\.pdf(\s|$)', line.strip(), flags=re.MULTILINE)
                if pdf:
                    tex = '{}.tex'.format(pdf.groups()[0])
                    if tex and make.parent.joinpath(tex).exists():
                        return make.parent / tex
        for name in MAIN_TEX_NAMES:
            if self.dir.joinpath(name).exists():
                return self.dir / name
        subdirs = [sd for sd in self.dir.iterdir() if sd.is_dir()]
        if len(subdirs) == 1 and not list(self.dir.glob('*.tex')):
            # All the main sources are nested in one directory.
            for name in MAIN_TEX_NAMES:
                if subdirs[0].joinpath(name).exists():
                    return subdirs[0] / name
        p = self._find_by_documentclass(self.dir)
        if p:
            return p
        for sd in subdirs:
            p = self._find_by_documentclass(sd)
            if p:
                return p


def includes_and_bib(d, main, chapterpath, no_bib):
    def norm_include(s):
        s = s.replace('\\chpath', chapterpath) \
            .replace('\\chapterpath', chapterpath) \
            .replace('\\folder', chapterpath)
        s = FIX_FILENAMES.get(s, s)
        if not s.endswith('.tex'):
            s += '.tex'
        return s

    tex = main.read_text(encoding='utf8')
    tex = tex.replace(r'\input{chapters/', r'\include{chapters/')
    tex = tex.replace(r'\input{phrasal-lfg-include', r'\include{phrasal-lfg-include')
    tex_lines = list(iter_texsoup_lines(main, tex=tex))
    for pincl in main.parent.glob('*-include.tex'):
        tex_lines.extend(iter_texsoup_lines(pincl))

    # Three variants of bib detection:
    bibs = [p for p in d.glob('*.bib')]
    if no_bib:  # explicitly marked as not having a bib
        bibs = []  # pragma: no cover
    elif len(bibs) == 1:  # only one choice for bibs
        pass  # pragma: no cover
    else:  # more bib files available, pick by parsing the main tex file
        bibs, bibnames = [], []
        for soup in tex_lines:
            if soup.bibliography or soup.addbibresource:
                bibnames.extend((soup.bibliography or soup.addbibresource).string.split(','))
        assert bibnames, str(main)
        for name in bibnames:
            if not name.endswith('.bib'):
                name += '.bib'
            bib = main.parent / name
            assert bib.exists(), str(bib)
            bibs.append(bib)
    includes = []

    for soup in tex_lines:
        incl = soup.include or soup.includechapter or soup.includepaper
        if incl:
            p = main.parent / norm_include(incl.string)
            # Allow case-insensitive matching:
            m = {pp.name.lower(): pp for pp in p.parent.glob('*.tex')}
            p = m.get(p.name.lower(), p)
            if not p.exists() and (p.stem in ['preface', 'acknowledgments']):
                continue
            if not p.exists() and p.stem == 'abbreviations':
                if p.parent.parent.joinpath('abbreviations.tex').exists():
                    p = p.parent.parent.joinpath('abbreviations.tex')
            assert p.exists(), str(p)
            includes.append(p)
    return includes, bibs



import pathlib
import collections

from tqdm import tqdm

from . import langsci
from . import glossa
from .base import Repository, Glottolog

assert langsci and glossa
PROVIDERS = {r.id: r for r in Repository.__subclasses__() if r.id}


def iter_publications(d='.', glottolog='glottolog', with_examples=False, exclude=None, **dirs):
    d = pathlib.Path(d)
    glottolog = Glottolog(glottolog)
    for rid, cls in PROVIDERS.items():
        if exclude and rid in exclude:
            continue  # pragma: no cover
        sd = dirs.get(rid, d / rid)
        if sd.exists():
            repos = cls(sd)
            glottolog.register_names(repos.lname_map)
            for pub in cls(sd).iter_publications():
                if with_examples:
                    for ex in pub.examples:
                        ex.Language_ID = glottolog(ex.Language_Name)
                        ex.Meta_Language_ID = glottolog(ex.Meta_Language_ID)
                        if not ex.Language_ID:
                            if ex.Comment in glottolog.by_name:  # pragma: no cover
                                ex.Language_ID = glottolog.by_name[ex.Comment].id
                                ex.Comment = None
                yield pub


def iter_examples(d='.', glottolog='glottolog', **dirs):  # pragma: no cover
    d = pathlib.Path(d)
    c = collections.Counter()
    glottolog = Glottolog(glottolog)
    for rid, cls in PROVIDERS.items():
        # if rid != 'glossa':
        if rid != 'langsci':
            continue
        sd = dirs.get(rid, d / rid)
        bibtex = sd / 'bibtex'
        if sd.exists():
            if not bibtex.exists():
                bibtex.mkdir()
            repos = cls(sd)
            glottolog.register_names(repos.lname_map)

            for pub in tqdm(cls(sd).iter_publications(), desc=rid):
                # if pub.record.int_id not in [6371]:
                # if rid == 'langsci' and pub.record.int_id < 289:
                #    continue

                # print(pub.id, len(pub.references), len(pub.cited))
                t = []
                for src in pub.cited_references:
                    #
                    # FIXME: add lgcode for refs related to identified examples!
                    #
                    t.append(src.bibtex())

                bibtex.joinpath(
                    '{}.bib'.format(pub.record.ID)).write_text('\n\n'.join(t), encoding='utf8')

                continue
                pid = '{}-{}'.format(rid, pub.record.ID)
                for i, ex in enumerate(pub.iter_examples()):
                    if ex.Language_ID is None:
                        ex.Language_ID = glottolog(ex.Language_Name)
                    print(ex.ID, ex.Language_ID)
                    c.update([pid])
    # for k, v in c.most_common(50):
    #    print(k, v)

    return

    d = pathlib.Path(d)
    # lsp = langsci.Repository(d / 'langsci')
    lsp = glossa.Repository(d / 'glossa')
    for pub in lsp.iter_publications():
        # if pub.record.int_id not in [94]:
        #    continue

        i = -1
        for i, ex in enumerate(pub.iter_examples()):
            print(ex.as_igt().pprint())
            break
        print('{}: {}'.format(pub.record.ID, i + 1))
        continue



from TexSoup import TexSoup


def iter_texsoup_lines(p, level=0, tex=None):
    """
    Yield TexSoup instances for each line in a tex file, including one level of input.
    """
    tex = tex or p.read_text(encoding='utf8')
    for line in tex.splitlines():
        try:
            soup = TexSoup(line.strip(), tolerance=1)
            if soup.input:
                if level == 0:
                    pp = p.parent / '{}.tex'.format(soup.input.string)
                    if pp.exists() and pp.stat().st_size < 300:
                        yield from iter_texsoup_lines(pp, level=level + 1)
            else:
                yield soup
        except:  # pragma: no cover # noqa: E722
            pass  # pragma: no cover



import typing
import pathlib
import collections

import attr
from pyigt import IGT
from pycldf.sources import Source
from clldutils.misc import lazyproperty
from pyglottolog import Glottolog as API

from linglit.util import clean_translation

__all__ = ['Glottolog', 'Record', 'Repository', 'Publication', 'Example']


class Glottolog:
    def __init__(self, glottolog: typing.Union[str, pathlib.Path, API]):
        if not isinstance(glottolog, API):  # pragma: no cover
            glottolog = API(glottolog)
        self.by_glottocode = {}
        self.by_isocode = {}
        self._by_name = {}
        self.by_custom_name = {}

        for lg in glottolog.languoids():
            self.by_glottocode[lg.id] = lg
            if lg.iso:
                self.by_isocode[lg.iso] = lg
            self._by_name[lg.name] = lg
            for _, names in lg.names.items():
                for name in names:
                    name = name.split('[')[0].strip()
                    if name not in self._by_name:
                        self._by_name[name] = lg

        self.by_name = {k: v for k, v in self._by_name.items()}
        self.api = glottolog

    def register_names(self, names: typing.Dict):
        self.by_name = {k: v for k, v in self._by_name.items()}
        for k, v in names.items():
            if v:
                v = self(v)
                if v:
                    self.by_name[k] = self.by_glottocode[v]

    def __call__(self, name: str) -> typing.Optional[str]:
        if name:
            if name in self.by_glottocode:
                return self.by_glottocode[name].id
            if name in self.by_isocode:
                return self.by_isocode[name].id
            gl = self.by_name.get(name)
            if gl:
                return gl.id


@attr.s
class Record:
    """
    Metadata record for a publication.
    """
    ID = attr.ib(validator=attr.validators.matches_re(r'^[0-9]+$'))
    DOI = attr.ib()
    metalanguage = attr.ib()
    objectlanguage = attr.ib()
    license = attr.ib()
    creators = attr.ib()
    title = attr.ib()
    year = attr.ib()

    def as_source(self) -> Source:  # pragma: no cover
        raise NotImplementedError()

    @property
    def has_open_license(self) -> bool:
        return self.license in ['CC-BY', 'CC-BY-SA']

    @property
    def current(self) -> bool:
        return True

    @property
    def int_id(self) -> int:
        return int(self.ID)


@attr.s
class Example:
    ID = attr.ib()
    Primary_Text = attr.ib()
    Analyzed_Word = attr.ib(validator=attr.validators.instance_of(list))
    Gloss = attr.ib(validator=attr.validators.instance_of(list))
    Translated_Text = attr.ib(converter=clean_translation)
    Language_Name = attr.ib()
    Comment = attr.ib(converter=lambda s: '; '.join(s) if isinstance(s, list) else s)
    Source = attr.ib(validator=attr.validators.instance_of(list))
    Language_ID = attr.ib(default=None)  # Assigned after initialization based on Language_Name
    Source_Path = attr.ib(default=None)
    Abbreviations = attr.ib(default=attr.Factory(collections.OrderedDict))
    Local_ID = attr.ib(default=None)
    Meta_Language_ID = attr.ib(default=None)

    def __str__(self):
        res = '({})'.format(self.Local_ID or self.ID)
        if self.Language_Name:
            res += ' {}'.format(self.Language_Name)
        if self.Source:
            res += ' ('
            for i, (sid, pages) in enumerate(self.Source):
                if i:
                    res += '; '
                res += sid
                if pages:
                    res += ': {}'.format(pages)
            res += ')'
        res += '\n'
        res += str(self.as_igt())
        return res

    def as_igt(self):
        return IGT(
            id=self.ID,
            phrase=self.Analyzed_Word,
            gloss=self.Gloss,
            translation=self.Translated_Text,
            abbrs=self.Abbreviations,
        )


class Publication:
    def __init__(self, record: Record, d: typing.Union[str, pathlib.Path], repos=None):
        self.record = record
        self.dir = pathlib.Path(d)
        self.repos = repos

    def __str__(self):
        return "{0.creators} {0.year}. {0.title}".format(self.record)

    @property
    def is_current(self) -> bool:
        return self.record.current

    @property
    def has_open_license(self) -> bool:
        return self.record.has_open_license

    @lazyproperty
    def cited_references(self) -> typing.List[Source]:
        return [ref for ref in self.references.values() if ref.id in self.cited]

    @lazyproperty
    def id(self) -> str:
        return '{}{}'.format(self.repos.id, self.record.ID)

    def as_source(self) -> Source:
        src = self.record.as_source()
        src.id = self.id
        return src

    @lazyproperty
    def references(self) -> typing.OrderedDict[str, Source]:
        res = collections.OrderedDict()
        for src in self.iter_references():
            sid = '{}:{}'.format(self.id, src.id)
            src['isreferencedby'] = self.id
            res[sid] = Source(src.genre, sid, _check_id=False, **src)
        return res

    def iter_references(self) -> typing.Generator[Source, None, None]:  # pragma: no cover
        raise NotImplementedError()

    @lazyproperty
    def cited(self) -> collections.Counter:
        res = collections.Counter()
        for key in self.iter_cited():
            res.update(['{}:{}'.format(self.id, key)])
        return res

    def iter_cited(self) -> typing.Generator[str, None, None]:  # pragma: no cover
        raise NotImplementedError()

    def example_sources(self, ex: Example) -> typing.List[Source]:
        """
        Resolve the source IDs for an example to `Source` instances.
        """
        return [self.references[sid] for sid, _ in ex.Source if sid != self.id] + [self.as_source()]

    @lazyproperty
    def examples(self):
        res = []
        for ex in self.iter_examples():
            ex.ID = '{}-{}'.format(self.id, ex.ID)
            refs = []
            for sid, pages in ex.Source:
                refs.append(('{}:{}'.format(self.id, sid), pages))
            if refs:
                pages = 'via'
                if ex.Local_ID:
                    pages += ':{}'.format(ex.Local_ID)
                refs.append((self.id, pages))
            else:
                refs.append((self.id, ex.Local_ID))
            ex.Source = refs
            res.append(ex)
        return res

    def iter_examples(self) -> typing.Generator[Example, None, None]:  # pragma: no cover
        raise NotImplementedError()


class Repository:
    id = None
    lname_map = {}

    def __init__(self, d):
        self.dir = pathlib.Path(d)

    def __getitem__(self, item: str) -> Publication:  # pragma: no cover
        raise NotImplementedError()

    def path(self, *comps):
        return self.dir.joinpath(*comps)

    def create(self, verbose=False):  # pragma: no cover
        raise NotImplementedError()

    def iter_publications(self):  # pragma: no cover
        raise NotImplementedError()



import re

from clldutils.text import replace_pattern

__all__ = ['read_tex']


def normalize_cite(t):
    def prepl(m):
        match = m.string[m.start():m.end()]
        for i, chunk in enumerate(match.split('{')):
            keys, _, _ = chunk.partition('}')
            if i:
                yield '\\cite{{{}}}'.format(keys)

    # Remove linebreaks in \cite* commands:
    t = re.sub(
        r'\\(cite[a-z]*)\n{',
        lambda m: '\\{}{{'.format(m.groups()[0]),
        t,
        flags=re.MULTILINE)

    t = replace_pattern(
        r"\\(parencites?|cites|textcites)\*?\s*(\([^)]*\)|\[[^]]*]|{[^}]+})+", prepl, t)

    t = t.replace('\\citesource{', '\\customcitesource{')

    # Replace starred commands with unstarred ones:
    t = re.sub(r'\\(cite[a-z]*)\*', lambda m: '\\{}'.format(m.groups()[0]), t)

    # Weird cases:
    t = t.replace(r'\citep[64–96],{Osborne2006}', r'\citep[64–96]{Osborne2006}')
    t = t.replace(r'\citetext{\citealp', r'{\citealp')

    # Now turn all the C|cite* variants into cite:
    t = re.sub(r'\\[Cc]ite[a-zA-Z]*\s*([\[{])', lambda m: r'\cite' + m.groups()[0], t)
    return t


def read_tex(p, with_input=True):
    """
    Read (and simplyfy) TeX from a file, resolving "input" commands.

    :param p:
    :param with_input:
    :return:
    """
    try:
        t = p.read_text(encoding='utf8')
    except UnicodeDecodeError:  # pragma: no cover
        t = p.read_text(encoding='latin1')

    t = normalize_cite(t)

    # For book 259:
    t = re.sub(
        r'\\input ([a-z]+|complex-predicates|control-raising)-include\.tex}'
        r'{\\input chapters/([a-z]+|complex-predicates|control-raising)-include\.tex}',
        lambda m: r'\input{chapters/%s-include.tex}}' % m.groups()[0],
        t
    )

    # Replace some weirdness/complexity that screws up example parsing:
    t = re.sub(r'\s*\\hfill\s*\[\\href{[^}]+}[^]]*]\s*$', '', t, re.MULTILINE)
    t = t.replace(r'\langinfobreak', r'\langinfo')  # Mostly a synonym.
    t = t.replace(r'{\db}{\db}{\db}', '[[[')
    t = t.replace(r'{\db}{\db}', '[[')
    t = t.replace(r'\textsc{id}:', '')
    t = t.replace(r'\glossN.', r'\glossN{}.')

    def repl_input(m):
        fname = m.groups()[1].strip()
        if not fname.endswith('.tex'):
            fname += '.tex'
        if not p.parent.joinpath(fname).exists() and '/' in fname:
            # look in the current directory:
            fname = fname.split('/')[-1]
        if p.parent.joinpath(fname).exists():
            yield '\n'
            yield read_tex(p.parent.joinpath(fname), with_input=False)
            yield '\n'

    if with_input:
        return replace_pattern(
            re.compile(r'^\s*\\(input|include)\s*{([^}]+)}', flags=re.MULTILINE), repl_input, t)
    return t



import re
import pathlib

PKG_PATH = pathlib.Path(__file__).parent
CFG_PATH = PKG_PATH / 'cfg'
STARTINGQUOTE = "`‘“"
ENDINGQUOTE = "'’”"
ELLIPSIS = '…'


def clean_translation(trs):
    trs = re.sub(r'\s+', ' ', trs.strip())
    try:
        if trs[0] in STARTINGQUOTE:
            trs = trs[1:]
        if trs[-1] in ENDINGQUOTE:
            trs = trs[:-1]
        if len(trs) > 1 and (trs[-2] in ENDINGQUOTE) and (trs[-1] == '.'):
            trs = trs[:-2] + '.'
        trs = trs.replace("()", "")
    except IndexError:  # s is  ''
        pass
    trs = trs.replace('...', ELLIPSIS)
    return trs



"""
We aggregate the relevant source files of langsci books in a repository.

This repository is created and synched with the upstream repositories by using the GitHub
REST API via [gh](https://cli.github.com/).
"""
import json
import base64
import pathlib
import subprocess

import attr
from clldutils.jsonlib import update_ordered, load
from clldutils.path import ensure_cmd

from linglit import base
from .catalog import Catalog, GITHUB_ORG
from .publication import Publication
from . import cfg

CATALOG_NAME = "catalog.tsv"
FILELIST_NAME = "files.json"
MISSING_TEX_SOURCES = [155, 192, 195, 255, 287, 297, 311, 325]
TEX_BRANCH = {187: 'master'}


@attr.s
class File:
    path = attr.ib(converter=pathlib.Path)  # "chapters/01.tex",
    sha = attr.ib()   # "9b724bc0398020c9c9b14926319a38d219f56786",
    mode = attr.ib()
    type = attr.ib()
    size = attr.ib()  # 23220,
    url = attr.ib()

    @property
    def content(self):
        data = gh_api(None, url=self.url)
        file_content = data['content']
        file_content_encoding = data.get('encoding')
        if file_content_encoding == 'base64':
            file_content = base64.b64decode(file_content)
        return file_content

    def same_content(self, p, shallow=True):
        if shallow:
            return p.stat().st_size == self.size
        return self.content == p.read_bytes()

    def fullpath(self, d):
        p = pathlib.Path(d).joinpath(self.path)
        p.parent.mkdir(parents=True, exist_ok=True)
        return p

    def save(self, d):
        self.fullpath(d).write_bytes(self.content)


def gh_api(item, path=None, url=None):
    if url is None:
        url = "https://api.github.com/repos/{}/{}{}".format(
            GITHUB_ORG, getattr(item, 'ID', item), path or '')
    print('requesting {}'.format(url))
    return json.loads(subprocess.check_output([ensure_cmd('gh'), "api", url]))


def branch_and_tree(item, olddata):
    default_branch = olddata[0] if olddata else gh_api(item)['default_branch']
    if int(getattr(item, 'ID', item)) in TEX_BRANCH:  # pragma: no cover
        default_branch = TEX_BRANCH[item.int_id]
    return default_branch, gh_api(item, '/git/trees/{}?recursive=1'.format(default_branch))


class Repository(base.Repository):
    id = 'langsci'
    lname_map = cfg.LNAME_MAP

    def __getitem__(self, item):
        return Publication(self.catalog[item], self.dir / item, self)

    def iter_publications(self):
        for item in self.catalog:
            # if item.int_id != 22:
            #     continue
            if item.int_id not in MISSING_TEX_SOURCES:
                yield Publication(item, self.dir / item.ID, self)

    def create(self, verbose=False):  # pragma: no cover
        """
        Create a repository from scratch (may need to be restarted a couple of times if
        GH rate limit problems are encountered):
        """
        self.fetch_catalog()
        self.fetch_filelist(refresh=True)
        self.fetch_files()

    @property
    def catalog(self):
        return Catalog.from_local(self.dir / CATALOG_NAME)

    def fetch_catalog(self):
        catalog = Catalog.from_remote()
        catalog.write(self.dir / CATALOG_NAME)

    def fetch_filelist(self, ids=None, refresh=False):
        ids = {int(i) for i in ids or []}
        with update_ordered(self.dir / FILELIST_NAME) as d:
            for item in self.catalog:
                if not refresh and (item.ID in d):
                    continue
                if not ids or (item.int_id in ids):
                    d[item.ID] = branch_and_tree(item, d.get(item.ID))

    def fetch_files(self, filelist=None):
        exclude = [
            'seriesinfo',
            'langsci/locale',
            'langsci-hyphenation',
            '.texpadtmp',
            'bibstyles.deprecated',
            'figures/',
            'graphics/',
            'biblatex-sp-unified',
            'draftinfo.tex',
            'generated/',
            'bibstyles/',
            'Figures/',
            'styles/tcb',
            '__MACOSX',
            'pdf/',
        ]
        for itemid, (_, filelist) in load(filelist or self.dir / FILELIST_NAME).items():
            sd = self.dir / itemid
            if not sd.exists():
                sd.mkdir()
            for file in filelist['tree']:
                if file['type'] not in ['tree', 'commit']:
                    file = File(**file)
                    if (file.path.suffix in ['.bib', '.tex'] or file.path.name == 'Makefile') \
                            and not any(e in str(file.path) for e in exclude):
                        fp = file.fullpath(sd)
                        if not fp.exists() or (not file.same_content(fp)):
                            file.save(sd)



import functools
import re
import typing
import hashlib

from pyigt.igt import IGT, NON_OVERT_ELEMENT
from pyigt.lgrmorphemes import MORPHEME_SEPARATORS
from TexSoup import TexSoup
from TexSoup.data import TexCmd

from linglit.base import Example, Publication
from .latex import to_text, strip_tex_comment

__all__ = ['iter_gll', 'make_example']

STARTINGQUOTE = "`‘"
ENDINGQUOTE = "'’"
EMPTY = NON_OVERT_ELEMENT
ELLIPSIS = '…'
CHAR_REPLS = {
    r'\v{s}': 'š',
    r"\'u": 'ú',
    r'\v{h}': "ȟ",
    r"\'a": "á",
}


def parse_langinfo(line):
    def get_name(arg):
        if len(arg.contents) == 1 and \
                isinstance(arg.contents[0], TexCmd) and \
                not arg.contents[0].args:
            return '\\' + arg.contents[0].name
        return ''.join(TexSoup(re.sub(r'\\label{[^}]+}', '', arg.string)).text)

    langinfo = TexSoup(r'\langinfo' + line.split(r'\langinfo')[-1], tolerance=1).langinfo
    if langinfo and len(langinfo.args) == 3:
        return (
            get_name(langinfo.args[0]),
            ''.join(TexSoup(langinfo.args[1].string).text),
            '_'.join(TexSoup(langinfo.args[2].string).text or ['']),
        )


def parse_cmd(cmd, line):
    try:
        # Cut out the command and its first argument from line:
        texcmd = '\\' + cmd + '{' + line.split('\\' + cmd + '{')[-1].split('}')[0]
        cmd = getattr(TexSoup(texcmd, tolerance=1), cmd)
    except:  # noqa: E722
        raise ValueError(line)  # pragma: no cover
    return (cmd.args[0].string.split('!')[-1], '', '')


parse_ili = functools.partial(parse_cmd, 'ili')
parse_il = functools.partial(parse_cmd, 'il')


def iter_gll(s):
    """
    Loop over the lines in a TeX file, detecting examples by matching start- and end-lines.

    FIXME: 123:
    \\syacex{Noun}{Pronoun}{984}
    {ܗܲܝܡܵܢܘܼܬ݂ܹܗ}
    {haymānut-ēh}
    {faith-\\poss.3\\masc}
    {his faith}
    {\\cite[70, \\S 91e]{MuraokaSyriac}}
    """
    gll_start = re.compile(r'\\(g[l]{2,3}|exg\.|ag\.|bg\.)([^a-zA-Z]|$)')
    glt_start = re.compile(r'\\(glt|trans|Transl|TranslMulti|rede)([^a-zA-Z]|$)')
    longexampleandlanguage_pattern = re.compile(r'\\\\}{([^}]+)}$')

    linfo = None
    ex_pattern = re.compile(
        r"\\ex\s+(?P<lname>[A-Z][a-z]+)\s+(\([A-Z][0-9],\s+)?\\cite[^{]+{(?P<ref>[^}]+)}")
    gll, in_gll, prevline, pregll = [], False, None, None
    for lineno, line in enumerate(s.split('\n')):
        line = strip_tex_comment(line).strip()
        if r'\langinfo' in line:
            res = parse_langinfo(line)
            if res:
                linfo = (res, lineno)
        elif r'\ili{' in line:
            res = parse_ili(line)
            if res:
                linfo = (res, lineno)
                line = line.replace('()', '').strip()
        elif r'\il{' in line:
            res = parse_il(line)
            if res:
                linfo = (res, lineno)
                line, rem = line.split(r'\il{', maxsplit=1)
                line += rem.split('}', maxsplit=1)[1] if '}' in rem else ''
                line = line.replace('()', '').strip()
        elif ex_pattern.match(line):
            m = ex_pattern.match(line)
            linfo = ((m.group('lname'), '', m.group('ref')), lineno)

        m = glt_start.search(line)
        if m:
            if gll and len(gll) < 10:
                #
                # We may need to fix the gloss line:
                mm = longexampleandlanguage_pattern.search(gll[-1])
                if mm:
                    linfo = ((mm.groups()[0], '', ''), lineno)
                    gll[-1] = gll[-1][:mm.start()]
                pre = line[:m.start()]
                line = line[m.end() - 1:]
                gll.append(pre)
                gll.append(line)
                # Return linfo it wasn't parsed too far from the example:
                yield linfo[0] if linfo and (lineno - linfo[1] < 25) else None, gll, pregll
            gll, in_gll = [], False
            continue
        m = gll_start.search(line)
        if m:
            line = line[m.end() - 1:]
            gll, pregll = [], prevline
            in_gll = True
        if in_gll:
            gll.append(line)

        prevline = line


def recombine(morphemes):
    """
    For better visual alignment, morphemes in examples are sometimes whitespace-separated - making
    them appear as words. This function recombines morphemes into words according to leading and
    trailing LGR morpheme separators.
    """
    chunk = []
    for c in morphemes:
        if not c:
            continue
        if c[0] in MORPHEME_SEPARATORS or (chunk and chunk[-1][-1] in MORPHEME_SEPARATORS):
            chunk.append(c)
        else:
            if chunk:
                yield ''.join(chunk)
            chunk = [c]
    if chunk:
        yield ''.join(chunk)


def fixed_alignment(pt, gl):
    """
    Given the aligned lines of an LGR IGT example, we try a couple of tricks to make sure both lines
    have the same number of word chunks.
    """
    comment = None

    # pre-process
    # Merge multi-word lexical glosses:
    multi_word_gloss = re.compile(r'(\s|^){([Ia-z ]+|North Wales|The birds)}(\s|,|$)')
    gl = multi_word_gloss.sub(
        lambda m: ' {} '.format(re.sub(r'\s+', '_', m.groups()[1])), gl).strip()
    gl = re.sub(r'(\s|^){}(\s|$)', ' _ ', gl.replace('{} {}', '{}  {}'))
    gl = re.sub(r'(\s|^){}{}(\s|$)', ' _ ', gl)
    gl = re.sub(r'(\s|^)~(\s|$)', ' _ ', gl)

    # Merge multi-word primary text groups:
    for k, v in {
        'Adnominal clause': '(Adnominal_clause)',
        'Adverbial clause': '(Adverbial_clause)',
        'Adverbila clause': '(Adverbial_clause)',

    }.items():
        pt = pt.replace(k, v)
    for k, v in CHAR_REPLS.items():
        pt = pt.replace(k, v)
    multi_word_pt = re.compile(r'(\s|^|-){([\w žąį./]+)}(\s|\.|$)')
    pt = multi_word_pt.sub(
        lambda m: ' {} '.format(re.sub(r'\s+', '_', m.groups()[1])), pt).strip()
    pt = re.sub(r'(\s|^){}(\s|$)', ' _ ', pt)

    # de-latex
    pt = to_text(pt)[0].split()
    gl = to_text(gl)[0].split()

    # post-process
    ellipsis_variants = [
        '....', '[…]', '[...]', '...', '[...].', '“…”', '[…].', '(...).', '…]']
    pt = [ELLIPSIS if w in ellipsis_variants else w for w in pt]
    gl = [ELLIPSIS if w in ellipsis_variants else w for w in gl]

    if len(pt) > len(gl):
        if pt[-1] == '.':
            pt = pt[:-1]
            pt[-1] += '.'
        elif pt[-1] == '[]':
            pt = pt[:-1]

    ldiff = len(pt) - len(gl)
    if ldiff == -1:
        if gl and gl[-1].startswith('[') and gl[-1].endswith(']'):
            comment = gl[-1][1:-1].strip()
            gl = gl[:-1]
    elif ldiff == 1:
        if ELLIPSIS in pt:
            gl.insert(pt.index(ELLIPSIS), ELLIPSIS)
        elif '/' in pt:
            gl.insert(pt.index('/'), '/')
        elif EMPTY in pt:
            gl.insert(pt.index(EMPTY), EMPTY)
        elif EMPTY + '.' in pt:
            gl.insert(pt.index(EMPTY + '.'), EMPTY)
        elif pt[-1] in [']', '].', ']?']:
            gl.append('_')
        elif re.fullmatch(r'\([^)]+\)', pt[-1]):
            comment = pt[-1].replace('(', '').replace(')', '')
            pt = pt[:-1]

    ldiff = len(pt) - len(gl)
    if ldiff > 0:
        if ldiff == pt.count(ELLIPSIS):
            for i, c in enumerate(pt):
                if c == ELLIPSIS:
                    gl.insert(i, ELLIPSIS)
        elif ldiff == pt.count('<') + pt.count('>'):
            for i, c in enumerate(pt):
                if c in ['<', '>']:
                    gl.insert(i, '_')

    pt_r, gl_r = list(recombine(pt)), list(recombine(gl))
    if len(pt_r) == len(gl_r):
        pt, gl = pt_r, gl_r
    return pt, gl, comment


def lines_and_comment(lines):
    """
    Figure out which lines of all lines between \\gll and \\glt to be considered as
    word/morpheme-segmented primary text and gloss.

    :param lines:
    :return:
    """
    res, comment, linfo = [], [], None
    for line in lines:
        line = line.strip()
        if line:
            try:
                s = TexSoup(line, tolerance=1)
                if s.jambox:
                    comment.append(s.jambox.string)
                    s.jambox.delete()
                    line = str(s)
            except:  # noqa: E722
                pass  # pragma: no cover
            if line:
                res.append(line)

    if len(res) > 2:
        # Language names as second argument to "longexampleandlanguage" commands.
        m = re.fullmatch(r'}{([A-Z][a-z]+)}', res[-1].split('\n')[0])
        if m:
            linfo = (m.groups()[0], '', '')
            res = res[:-1]
        else:
            # Comments in square brackets appended to the example.
            m = re.fullmatch(r'\[([\w ]+)]', to_text(res[-1].split('\n')[0])[0].strip())
            if m:
                comment.append(m.groups()[0])
                res = res[:-1]
            else:
                # A single word: Considered a comment.
                m = re.fullmatch(r'([\w]+)', to_text(res[-1].split('\n')[0])[0].strip())
                if m:
                    comment.append(m.groups()[0])
                    res = res[:-1]
                else:
                    # Language names appended as special comment in parentheses to the example.
                    m = re.fullmatch(
                        r'\(?([A-Z][a-z]+(-English)?|[0-9/]+|[A-Z][A-Z]+)\)?',
                        to_text(res[-1].split('\n')[0])[0].strip())
                    if m:
                        if m.groups()[0][0].isalpha() and m.groups()[0][0].islower():
                            linfo = (m.groups()[0], '', '')
                        else:
                            comment.append(m.groups()[0])
                        res = res[:-1]
    return [r.replace('\n', ' ') for r in res], '; '.join(comment), linfo


def make_example(
        pub: Publication,
        linfo: typing.Tuple[str, str, str],
        gll: typing.List[str],
        prevline: str) -> typing.Optional[Example]:
    _, _, refs = to_text(prevline)
    comment = []
    if linfo:
        linfo = [to_text(s or '')[0] for s in linfo]
        if linfo[2]:
            comment.append(linfo[2])

    lname = linfo[0] if linfo and linfo[0] else None

    #
    # At this point `gll` is just a bunch of text lines containing latex formatting.
    #
    # We assume the last text line to be the translation ...
    aligned, translation = '\n'.join(gll[:-1]), gll[-1]
    translation, cmt, _refs = to_text(translation)
    if _refs:
        refs.extend(_refs)
    if cmt:
        comment.append(cmt)

    # ... and split the remainder at latex newlines:
    aligned = [line.strip() for line in re.split(r'\\(?:\\|newline)', aligned) if line.strip()]

    # book-specifics:
    if pub.record.int_id == 212:
        if len(aligned) > 2:
            if 'footnotesize' in aligned[2]:
                aligned = aligned[:2]

    aligned, cmt, linfo2 = lines_and_comment(aligned)
    if linfo2 and linfo2[0]:
        lname = linfo2[0]
    if cmt:
        comment.append(cmt)

    al = []
    for line in aligned:
        delatexed, cmt, _refs = to_text(line)
        if _refs:
            refs.extend(_refs)
        if cmt:
            comment.append(cmt)
        if delatexed.replace('*', '').strip():  # Not an empty line (when texed).
            al.append(line)
    aligned = al

    if len(aligned) == 3:
        # There's a separate line for the morpheme-segmented primary text!
        obj, pt, gl = aligned
    elif len(aligned) == 2:
        pt, gl = aligned
        obj = None
    elif len(aligned):
        if len(aligned) == 4 and aligned[3].startswith(r'}\\jambox'):
            obj, pt = aligned[0], aligned[1]
            gl = aligned[2] + aligned[3]
        else:  # Dunno what to do here ...
            # print(filename)
            # print(len(aligned), aligned)
            # print('---')
            return
    else:  # ... or here.
        return
    if obj:
        obj, cmt, _refs = to_text(obj)
        if _refs:
            refs.extend(_refs)
        if cmt:
            comment.append(cmt)

    pt, gl, cmt = fixed_alignment(pt, gl)
    if cmt:
        comment.append(cmt)
    if len(pt) != len(gl):
        if gl and gl[-1] in ['()', '*()']:
            gl = gl[:-1]
    if len(pt) != len(gl):
        return

    obj = obj or IGT(phrase=pt, gloss=gl).primary_text
    return Example(
        ID=hashlib.sha256(obj.replace('.', '').encode('utf8')).hexdigest()[:10],
        Primary_Text=obj,
        Analyzed_Word=pt,
        Gloss=gl,
        Translated_Text=translation,
        Language_Name=lname if len(lname or '') > 2 else None,
        Comment=comment,
        Source=refs,
    )



"""
Create a bibliography by merging all glossa publications in the repository and their references.
"""
from clldutils.path import TemporaryDirectory

from linglit.cli_util import add_provider, get_provider
from linglit.bibtex import iter_entries, iter_merged


def register(parser):
    add_provider(parser)


def run(args):
    def bibtex(src):
        return '{}\n'.format(src.bibtex())

    repos = get_provider(args)
    with TemporaryDirectory() as tmp:
        for pub in repos.iter_publications():
            with tmp.joinpath('{}.bib'.format(pub.id)).open('w') as bib:
                bib.write(bibtex(pub.as_source()))
                for src in pub.cited_references:
                    bib.write(bibtex(src))
        for src, _ in iter_merged(iter_entries(tmp)):
            print(bibtex(src))



import re
import functools

from clldutils.misc import nfilter
from lxml.etree import fromstring, tostring
from pyigt import IGT
from pycldf.sources import Source


def element(s):
    if isinstance(s, str):
        s = fromstring(s)
    return s


def parse(p):
    return fromstring(p.read_bytes().replace(b'&nbsp;', b'&#160;'))


def translate(in_, out_, s):
    tr = dict(zip(in_, out_))
    try:
        return ''.join(tr[c] for c in s)
    except (KeyError, TypeError):
        return ''


sup = functools.partial(
    translate,
    "ABCDEFGHIJKLMNOPRSTUVWXYZabcdefghijklmnoprstuvwxyz0123456789+-=()",
    "ᴬᴮᶜᴰᴱᶠᴳᴴᴵᴶᴷᴸᴹᴺᴼᴾᴿˢᵀᵁⱽᵂˣʸᶻᵃᵇᶜᵈᵉᶠᵍʰᶦʲᵏˡᵐⁿᵒᵖʳˢᵗᵘᵛʷˣʸᶻ⁰¹²³⁴⁵⁶⁷⁸⁹⁺⁻⁼⁽⁾")
sub = functools.partial(translate, "aeox0123456789+-=()", "ₐₑₒₓ₀₁₂₃₄₅₆₇₈₉₊₋₌₍₎")


def iter_text(p, strict=True):
    for e in p.xpath('child::node()'):
        if getattr(e, 'tag', None):
            if e.tag == 'sc':
                yield text(e, strict=strict).upper()
            elif e.tag == 'italic':
                yield text(e, strict=strict)
            elif e.tag == 'sup':
                yield sup(e.text)
            elif e.tag == 'sub':
                yield sub(e.text)
            elif e.tag == 'upper':
                if e.text:
                    yield e.text.upper()
                else:  # pragma: no cover
                    raise ValueError(tostring(e))
            elif e.tag in ['xref', 'ext-link']:
                yield e.text
            elif e.tag in ['table-wrap', 'disp-formula']:
                pass
            elif e.tag == 'list':
                # FIXME: '<list xmlns:mml="http://www.w3.org/1998/Math/MathML"
                #  xmlns:xlink="http://www.w3.org/1999/xlink" xmlns:xsi="http://www.w3.org/
                #  2001/XMLSchema-instance" list-type="alpha-lower">\n<list-item>
                #  <p>&#8220;Shark caught fish, but penguin did not catch fish.&#8221;
                #  </p></list-item>\n<list-item><p>&#8220;Shark caught fish, but shark
                #  did not catch penguin.&#8221;</p></list-item>\n<list-item><p>Both
                #  (a) and (b)</p></list-item>\n<list-item><p>I am not sure</p>
                #  </list-item>\n</list>'
                pass  # pragma: no cover
            elif e.text:
                if strict:
                    assert not isinstance(e.tag, str) or (e.tag in [
                        'bold', 'italic', 'underline', 'strike', 'monospace', 'inline-formula'
                    ]), tostring(e)
                yield e.text
        else:
            yield str(e)


def parse_citation(s):
    refs = []
    s = element(s)
    sub = s.xpath('p/sub')
    if sub and sub[0].xpath('xref'):
        for xref in sub[0].xpath('xref'):
            refs.append((xref.get('rid'), xref.get('ref-type'), xref.text))
        return text(sub[0]), refs


def text(e, strict=True):
    return ''.join(nfilter(iter_text(element(e), strict=strict)))


def t(s, multi=False):
    p = s.xpath('p')
    if not multi:
        assert len(p) == 1 or p[1].xpath('table-wrap') or (
                len(p) == 2 and  # noqa: W504
                text(p[1]).strip() in ['', '↔', 'ɛ elsewhere', 'ə elsewhere']), \
            '\n'.join(tostring(pp).decode('utf8') for pp in p)
        return text(p[0])
    return '\n'.join(text(pp) for pp in p)


def parse_igt(d):
    """
    #
    # FIXME: orthogonal layout - see tests/glossa/6371.xml
    #
    """
    comment, refs = [], []
    comment_pattern = re.compile(r'\s\s+\(([^)]+)\)$')
    aw, gl, tr = [], [], []
    for i, li in enumerate(d.xpath('list-item')):
        words = li.xpath("list[@list-type='word']")
        if words:
            for w in words:
                tiers = w.xpath('list-item')
                if len(tiers) != 2:
                    # We don't know how to handle alignments with more than two lines!
                    #
                    # FIXME: Detect orthogonal layout, where first (nth) word holds list words,
                    # second (n+1th) word holds list of glosses, enclosed in &#8216; and &#8217;
                    # See 4835
                    #
                    return
                word = t(tiers[0])
                m = comment_pattern.search(word)
                if m:
                    comment.append(m.groups()[0])
                    word = word[:m.start()]
                aw.append(word)
                gl.append(t(tiers[1]))
        fs = li.xpath("list[@list-type='final-sentence']")
        if fs:
            for i, l in enumerate(fs):
                items = l.xpath('list-item')
                for j, lii in enumerate(items):
                    res = parse_citation(lii)
                    if res:
                        comment.append(res[0])
                        refs = res[1]
                    tr.append(t(lii, multi=True))
                break

    return aw, gl, '\n'.join(tr), '; '.join(comment), refs


def iter_igt(d, abbrs):
    seen, count, number, letter = set(), 0, None, None
    lang, refs = None, []
    for gloss in d.xpath(".//list[@list-type='gloss']"):
        numbers = [
            t(li.xpath('list-item')[0])
            for li in gloss.xpath(".//list[@list-type='wordfirst']")]
        for n in numbers:
            m = re.match(r'\(([0-9]+|[iv]+)\)', n)
            if m:
                nn = m.groups()[0]
                if number:
                    letter = None
                    lang, refs = None, []
                number = nn
            m = re.fullmatch(r'([a-z])\.', n)
            if m:
                letter = m.groups()[0]
                break

        for ll in gloss.xpath("list-item/list[@list-type='sentence-gloss']"):
            if ll.xpath(".//list[@list-type='gloss']"):
                # there are nested examples! skip the wrapper.
                continue  # pragma: no cover
            # look for language and refs:
            fs = ll.xpath("list-item/list[@list-type='final-sentence']")
            if fs:
                items = fs[0].xpath('list-item')
                if items and len(items) == 1:
                    lname = parse_language_name(items[0])
                    if lname and len(lname) > 1 and lname[0].isupper():
                        lang = lname
                    for xref in items[0].xpath('p/xref'):
                        refs.append((xref.get('rid'), xref.get('ref-type'), xref.text))

            if ll.xpath("list-item/list[@list-type='word']") and not ll.xpath(".//inline-graphic"):
                try:
                    res = parse_igt(ll)
                except AssertionError:  # pragma: no cover
                    res = None
                if res:
                    igt = IGT(phrase=res[0], gloss=res[1], translation=res[2], abbrs=abbrs)
                    if igt.primary_text not in seen:
                        refs.extend(res[4])
                        count += 1
                        yield count, number, letter, lang, refs, igt, res[3]
                        seen.add(igt.primary_text)


def parse_language_name(e):
    e = element(e)
    n = None
    if e.xpath('p/italic'):
        if text(e.xpath('p')[0]).startswith('‘'):
            return
        n = e.xpath('p/italic')[0].text
    if e.xpath('p') and re.match(
            r'([A-Z][a-z]+)(\s+[A-Z][a-z]+)*\s+\(', e.xpath('p')[0].text or ''):
        n = e.xpath('p')[0].text.split('(')[0].strip()
    if e.xpath('p/sub') and not (e.xpath('p')[0].text or '').strip():
        n = text(e.xpath('p/sub')[0])
    if e.xpath('p/sc') and not (e.xpath('p')[0].text or '').strip():
        n = text(e.xpath('p/sc')[0])
    if n:
        return ' '.join(w.capitalize() for w in n.split())


def names(xp):
    res = []
    for n in xp:
        try:
            if not n.xpath('surname'):
                continue  # pragma: no cover
            name = n.xpath('surname')[0].text
            gn = n.xpath('given-names')
            if gn:
                name += ', {}'.format(gn[0].text)
            res.append(name)
        except:  # pragma: no cover # noqa: E722
            raise ValueError(tostring(n))  # pragma: no cover
    return ' and '.join(res)


def abbreviations(doc):
    sec = doc.xpath(".//sec[title/text()='Abbreviations']")
    res = {}
    if sec:
        ps = sec[0].xpath('p')
        if not ps:  # pragma: no cover
            return res
        abbr, desc = None, []
        for p in ps:
            for e in p.xpath('child::node()'):
                if getattr(e, 'tag', None):
                    if e.tag == 'sc':
                        if abbr:
                            res[abbr] = ''.join(desc)
                        abbr, desc = e.text.upper(), []
                    elif e.text:
                        desc.append(e.text)
                else:
                    desc.append(str(e))
        if abbr:
            res[abbr] = ''.join(desc)

    def clean_desc(s):
        s = s.strip()
        if s.startswith('='):
            s = s[1:].strip()
        if s.endswith(','):
            s = s[:-1].strip()
        return s
    return {k: clean_desc(v) for k, v in res.items()}


def metadata(p, doc):
    """
    """
    license = doc.xpath('.//license')[0].get('{http://www.w3.org/1999/xlink}href')
    assert 'creativecommons.org/licenses/by/' in license, license
    doi = doc.xpath(".//article-id[@pub-id-type='doi']")[0].text
    assert doi
    title = re.sub(r'\s+', ' ', text(doc.xpath(".//article-title")[0]))
    assert title

    return dict(
        ID=p.stem,
        DOI=doi,
        metalanguage='eng',
        objectlanguage=False,
        license=license,
        creators=names(doc.xpath(".//contrib[@contrib-type='author']/name")),
        title=title,
        year=doc.xpath(".//pub-date/year")[0].text,
        doc=doc,
    )


def parse_ref(ref):
    ref = element(ref)
    sid = ref.get('id')
    mixed = False
    try:
        ref = ref.xpath('element-citation')[0]
    except IndexError:
        mixed = True
        ref = ref.xpath('mixed-citation')[0]

    genre = {
        'journal': 'article',
        'confproc': 'inproceedings',
        'webpage': 'online',
        '/label': 'misc',
        'conf-proc': 'inproceedings',
        'cofproc': 'inproceedings',
    }.get(ref.get('publication-type'), ref.get('publication-type'))
    md = {}
    for pg in ref.xpath('person-group'):
        md[pg.get('person-group-type')] = names(pg.xpath('name'))
    if ref.xpath('string-name'):
        md['author'] = names(ref.xpath('string-name'))
    for year in ref.xpath('year'):
        md['year'] = year.text
        break
    for year in ref.xpath('edition'):
        md['edition'] = year.text
        break
    for uri in ref.xpath('uri'):
        md['url'] = uri.text
        break
    if ref.xpath('fpage'):
        md['pages'] = ref.xpath('fpage')[0].text
        lpage = ref.xpath('lpage')
        if lpage:
            md['pages'] += '-{}'.format(lpage[0].text)
    p = ref.xpath("pub-id[@pub-id-type='doi']")
    if p:
        md['doi'] = p[0].text
    p = ref.xpath("article-title")
    if p:
        md['title'] = text(p[0])
    p = ref.xpath("chapter-title")
    if p:
        genre = 'incollection'
        md['title'] = text(p[0])
    p = ref.xpath('source')
    if p:
        f = {
            'incollection': 'booktitle',
            'article': 'journal',
        }.get(genre, 'title')
        md[f] = text(p[0])
        if md[f].endswith('. doctoral dissertation'):
            md[f] = md[f].replace('. doctoral dissertation', '')
            genre = 'phdthesis'
    if mixed and '. doctoral dissertation' in text(ref, strict=False):
        genre = "phdthesis"
    if mixed and 'MA dissertation' in text(ref, strict=False):  # pragma: no cover
        genre = "mastersthesis"
    for xp, field in [
        ('conf-loc', 'address'),
        ('conf-name', 'howpublished'),
        ('conf-sponsor', 'publisher'),
        ('publisher-loc', 'address'),
        ('publisher-name', 'publisher'),
        ('volume', 'volume'),
        ('issue', 'number'),
        ('issn', 'issn'),
        ('isbn', 'isbn'),
    ]:
        p = ref.xpath(xp)
        if p:
            md[field] = text(p[0])

    if ('thesis' in genre) and ('publisher' in md):
        md['school'] = md.pop('publisher')

    return Source(genre, sid, **md)


def refs(doc):
    for ref in doc.xpath('.//ref-list/ref'):
        yield parse_ref(ref)



"""
Update a linglit data repository
"""
from linglit.cli_util import add_provider, get_provider


def register(parser):
    add_provider(parser)


def run(args):  # pragma: no cover
    repo = get_provider(args)
    repo.create(verbose=True)



import io
import urllib.request

import attr
from csvw.dsv import reader, UnicodeWriter
from pycldf.sources import Source

from linglit import base

URL = "https://raw.githubusercontent.com/langsci/opendata/master/catalog.csv"
GITHUB_ORG = 'langsci'


@attr.s
class Record(base.Record):
    edited = attr.ib()
    superseded = attr.ib()
    pages = attr.ib()
    series = attr.ib()
    seriesnumber = attr.ib()

    @property
    def current(self):
        return self.superseded != 'True'

    def as_source(self):
        md = dict(
            title=self.title,
            year=self.year,
            series=self.series,
            number=self.seriesnumber,
            address="Berlin",
            publisher="Language Science Press",
            doi=self.DOI,
        )
        parts = [ss.strip() for ss in self.creators.replace(' & ', ',').split(',')]
        assert len(parts) % 2 == 0, 'Odd number of commas: {}'.format(self.creators)
        names = ' and '.join(['{}, {}'.format(*parts[i:i + 2]) for i in range(0, len(parts), 2)])
        names = names.replace('\u180e', '').strip()
        if names.endswith(','):
            names = names[:-1].strip()
        if self.edited:
            md['editor'] = names
        else:
            md['author'] = names
        return Source('book', self.ID, **md)


class Catalog:
    def __init__(self, rows):
        self.items = [Record(**row) for row in rows]

    @classmethod
    def from_remote(cls):
        return cls(reader(
            io.StringIO(urllib.request.urlopen(URL).read().decode('utf8')),
            dicts=True, delimiter='\t'))

    @classmethod
    def from_local(cls, p):
        return cls(reader(p, dicts=True, delimiter='\t'))

    def write(self, dest):
        with UnicodeWriter(dest, delimiter='\t') as w:
            w.writerow([f.name for f in attr.fields(Record)])
            for item in self.items:
                w.writerow(attr.astuple(item))

    def __len__(self):
        return len(self.items)

    def __iter__(self):
        for item in self.items:
            if item.has_open_license:
                yield item

    def __getitem__(self, item):
        for it in self.items:
            if it.ID == item:
                return it
        else:
            raise KeyError(item)
